#!/usr/bin/env python
# -*- coding: utf-8 -*-

import json
import os
import re
from collections import OrderedDict, defaultdict
from copy import deepcopy
from typing import Type

import dill
import numpy as np
import torch
import torch.nn as nn
import torch.nn.functional as F
from loguru import logger
from torch.nn import CrossEntropyLoss, MSELoss
from tqdm import tqdm

from transformers import AutoModel, BertModel, BertPreTrainedModel

from ...modules import FocalLoss, LabelSmoothingCrossEntropy
from ..arguments import (DataArguments, ModelArguments, TaskArguments,
                         TrainingArguments,
                         generate_method_kwargs_from_arguments)
from ..data.samples import GlueSamples
#  from ...losses import DiceLoss, FocalLoss
#  from .ner_decodes import crf_decode, mrc_decode, span_decode
from .task import BaseDataset, BaseTask, TaskData, TaskRunner, TransformerModel

#  def softmax(x):
#      e_x = np.exp(x - np.max(x, axis=1).reshape(-1, 1))
#      return e_x / e_x.sum(axis=1).reshape(-1, 1)


def get_p_r_f1(tp, fp, fn):
    p = tp / (tp + fp) if tp + fp != 0 else 0
    r = tp / (tp + fn) if tp + fn != 0 else 0
    f1 = 2 * p * r / (p + r) if p + r != 0 else 0
    return np.array([p, r, f1])


def generate_char2token(offset_mapping, num_text_len):
    char2token = [-1] * num_text_len

    for i, (start, end) in enumerate(offset_mapping):
        if start == 0 and end == 0:
            continue
        #  token2char[i] = start
        for j in range(start, end):
            char2token[j] = i

    return char2token


def generate_token2chars(offset_mapping):
    token2char = [-1] * (len(offset_mapping) + 1)

    for i, (start, end) in enumerate(offset_mapping):
        if start == 0 and end == 0:
            token2char[i] = -1
            continue
        token2char[i] = start
        token2char[i + 1] = end

    return token2char


# ------------------------------ Dataset ------------------------------
class NerDataset(BaseDataset):
    """
    完成模型输入数据的编码工作
    """
    def __init__(self, *args, **kwargs):
        super(NerDataset, self).__init__(*args, **kwargs)

    def _encode_item(self, x):
        guid, text, _, tags = x

        #  logger.info(f"{guid}, {text}, tags: {tags}")

        def seg_encode(text, tags):
            batch_texts = [text]
            batch_tags = [tags]
            # -------- input_ids, attention_mask, token_type_ids --------
            encodings = self.tokenizer.batch_encode_plus(
                batch_texts,
                padding=self.data_args.padding,
                max_length=self.data_args.max_length,
                add_special_tokens=True,
                truncation=True,
                return_offsets_mapping=True)

            batch_input_ids = torch.from_numpy(
                np.array(encodings.input_ids, dtype=np.int64))
            batch_attention_mask = torch.from_numpy(
                np.array(encodings.attention_mask, dtype=np.int64))
            batch_token_type_ids = torch.from_numpy(
                np.array(encodings.token_type_ids, dtype=np.int64))

            batch_offset_mapping = encodings.offset_mapping
            batch_tokens = [
                #
                [text[b:e] for b, e in offset_mapping if b > 0 or e > 0]
                #
                for text, offset_mapping in zip(batch_texts,
                                                batch_offset_mapping)
            ]
            batch_char2tokens = [
                generate_char2token(offset_mapping,
                                    len(text)) for offset_mapping, text in zip(
                                        batch_offset_mapping, batch_texts)
            ]
            batch_token2chars = [
                generate_token2chars(offset_mapping)
                for offset_mapping in batch_offset_mapping
            ]

            #  for text, offset_mapping in zip(batch_texts, batch_offset_mapping):
            #      tokens = [
            #          text[b:e] for b, e in offset_mapping if b > 0 or e > 0
            #      ]

            input_ids = batch_input_ids[0]
            attention_mask = batch_attention_mask[0]
            token_type_ids = batch_token_type_ids[0]
            tokens = batch_tokens[0]
            char2token = batch_char2tokens[0]
            token2char = batch_token2chars[0]

            # -------- labels --------
            if tags is not None:
                num_tokens = input_ids.shape[0]
                start_ids = [0] * num_tokens
                end_ids = [0] * num_tokens
                for tag in tags:
                    c = tag['category']
                    s = tag['start']
                    m = tag['mention']
                    e = s + len(m) - 1

                    #  logger.info(f"text: {text}")
                    #  logger.info(f"tokens: {tokens}")
                    #  logger.info(f"tag: {tag}")
                    #  logger.info(f"s: {s}, e: {e}, char2token: {char2token}")
                    s = char2token[s]
                    if s < 1:
                        logger.warning(f"{guid}, {text}, tags: {tags}")
                    assert s >= 1
                    e = char2token[e]
                    if e < 1:
                        logger.warning(f"{guid}, {text}, tags: {tags}")
                    assert e >= 1
                    #  logger.debug(
                    #      f"num_tokens: {num_tokens}, s: {s}, e: {e}, c: {c}, m: {m}, text: {text}"
                    #  )
                    c_id = self.label2id[c]
                    #  assert c_id < 10, f"c_id: {c_id}"
                    #  logger.info(f"c_id: {c_id}")
                    start_ids[s] = c_id
                    end_ids[e] = c_id

                start_ids = torch.from_numpy(
                    np.array([start_ids], dtype=np.int64))[0]
                end_ids = torch.from_numpy(np.array([end_ids],
                                                    dtype=np.int64))[0]
            else:
                start_ids = None
                end_ids = None

            return {
                'tokens': tokens,
                'char2token': char2token,
                'token2char': token2char,
                'input_ids': input_ids,
                'attention_mask': attention_mask,
                'token_type_ids': token_type_ids,
                'start_ids': start_ids,
                'end_ids': end_ids,
                'tags': tags
            }

        all_encoded = []
        seg_len = self.data_args.max_length - 2
        seg_stride = int(seg_len / 2)
        seg_offset = 0
        while seg_offset < len(text):
            s_seg = seg_offset
            e_seg = seg_offset + seg_len - 1
            seg_text = text[s_seg:e_seg + 1]
            if len(seg_text) == 0:
                continue
            e_seg = min(e_seg, len(seg_text) - 1)
            self.seg_spans.append((guid, s_seg, e_seg))

            if tags is not None:
                seg_tags = []
                if tags:
                    for tag in tags:
                        s = tag['start']
                        e = s + len(tag['mention']) - 1
                        if s >= s_seg and e <= e_seg:
                            seg_tag = deepcopy(tag)
                            seg_tag['start'] -= seg_offset
                            seg_tags.append(seg_tag)
                            #  logger.info(
                            #      f"text: {text}\n"
                            #      f"tag: {tag}\n"
                            #      f"seg_len: {seg_len}, seg_stride: {seg_stride}\n"
                            #      f"seg_offset: {seg_offset}\n"
                            #      f"seg_tag: {seg_tag}\n")
                            assert (
                                seg_text[s:e + 1] == tag['mention'],
                                f"text: {text}\n"
                                f"tag: {tag}\n"
                                f"seg_len: {seg_len}, seg_stride: {seg_stride}\n"
                                f"seg_offset: {seg_offset}\n"
                                f"seg_tag: {seg_tag}\n")

                if seg_tags:
                    encoded = seg_encode(seg_text, seg_tags)
                    all_encoded.append(encoded)
                else:
                    if self.data_args.preserve_no_entity:
                        encoded = seg_encode(seg_text, [])
                        all_encoded.append(encoded)
            else:
                encoded = seg_encode(seg_text, None)
                all_encoded.append(encoded)

            seg_offset += seg_stride

        return all_encoded

    @classmethod
    def collate_fn(cls, batch):
        stacked_batch = {}
        #  for key in [
        #          'input_ids', 'attention_mask', 'token_type_ids', 'start_ids',
        #          'end_ids', 'tags'
        #  ]:
        #      key_batch = [e[key] for e in batch if e[key] is not None]
        #      if key_batch:
        #          if key == 'tags':
        #              stacked_batch['tags'] = key_batch
        #          else:
        #              #  if key == 'start_ids':
        #              #      logger.warning(f"key: {key}, key_batch: {key_batch}")
        #              batch_values = torch.stack(key_batch)
        #              stacked_batch[key] = batch_values
        #      else:
        #          stacked_batch[key] = None

        not_none_tensor_keys = [
            'input_ids', 'attention_mask', 'token_type_ids'
        ]
        maybe_none_tensor_keys = ['start_ids', 'end_ids']
        not_tensor_keys = ['tokens', 'char2token', 'token2char', 'tags']

        # not None tensors
        for key in not_none_tensor_keys:
            key_batch = [e[key] for e in batch if e[key] is not None]
            batch_values = torch.stack(key_batch)
            stacked_batch[key] = batch_values
        # maybe None tensors
        for key in maybe_none_tensor_keys:
            key_batch = [e[key] for e in batch if e[key] is not None]
            if key_batch:
                batch_values = torch.stack(key_batch)
                stacked_batch[key] = batch_values
            else:
                stacked_batch[key] = None
        # not tensors
        for key in not_tensor_keys:
            key_batch = [e[key] for e in batch if e[key] is not None]
            stacked_batch[key] = key_batch

        return stacked_batch


# ------------------------------ TaskData ------------------------------
class NerData(TaskData):
    """
    指定任务专属的Dataset类型，并提供训练、验证、测试集实例。
    """
    def __init__(self, *args, **kwargs):
        super(NerData, self).__init__(*args, **kwargs)

    def build_train_dataset(self):
        return NerDataset(self.data_args, self._splitted_train_samples.rows,
                          self.label2id, self.tokenizer)

    def build_val_dataset(self):
        return NerDataset(self.data_args, self._splitted_val_samples.rows,
                          self.label2id, self.tokenizer)

    def build_test_dataset(self):
        return NerDataset(self.data_args, self.test_samples.rows,
                          self.label2id, self.tokenizer)


# ------------------------------ SpanModel ------------------------------
class FeedForwardNetwork(nn.Module):
    def __init__(self, input_size, hidden_size, output_size, dropout_rate=0):
        super(FeedForwardNetwork, self).__init__()
        self.dropout_rate = dropout_rate
        self.linear1 = nn.Linear(input_size, hidden_size)
        self.linear2 = nn.Linear(hidden_size, output_size)

    def forward(self, x):
        x_proj = F.dropout(F.relu(self.linear1(x)),
                           p=self.dropout_rate,
                           training=self.training)
        x_proj = self.linear2(x_proj)
        return x_proj


class PoolerStartLogits(nn.Module):
    def __init__(self, hidden_size, num_classes):
        super(PoolerStartLogits, self).__init__()
        self.dense = nn.Linear(hidden_size, num_classes)

    def forward(self, hidden_states, p_mask=None):
        x = self.dense(hidden_states)
        return x


class PoolerEndLogits(nn.Module):
    def __init__(self, hidden_size, num_classes):
        super(PoolerEndLogits, self).__init__()
        self.dense_0 = nn.Linear(hidden_size, hidden_size)
        self.activation = nn.Tanh()
        self.LayerNorm = nn.LayerNorm(hidden_size)
        self.dense_1 = nn.Linear(hidden_size, num_classes)

    def forward(self, hidden_states, start_ids=None, p_mask=None):
        x = self.dense_0(torch.cat([hidden_states, start_ids], dim=-1))
        x = self.activation(x)
        x = self.LayerNorm(x)
        x = self.dense_1(x)
        return x


class BertSpanForNer(BertPreTrainedModel):
    def __init__(self, config):
        super(BertSpanForNer, self).__init__(config)
        self.soft_label = False  #config.soft_label
        self.num_labels = config.num_labels
        #  self.loss_type = "CrossEntropyLoss"  #"FocalLoss"  #config.loss_type
        self.loss_type = "FocalLoss"
        self.focalloss_gamma = 2.0  #config.focalloss_gamma
        self.focalloss_alpha = 0.25  #config.focalloss_alpha
        #  self.diceloss_weight = config.diceloss_weight
        logger.warning(f"BertSpanForNer.num_labels: {self.num_labels}")

        self.bert = BertModel(config)
        self.dropout = nn.Dropout(config.hidden_dropout_prob)
        self.start_fc = PoolerStartLogits(config.hidden_size, self.num_labels)
        if self.soft_label:
            self.end_fc = PoolerEndLogits(config.hidden_size + self.num_labels,
                                          self.num_labels)
        else:
            self.end_fc = PoolerEndLogits(config.hidden_size + 1,
                                          self.num_labels)
        self.init_weights()

    def forward(self,
                input_ids,
                attention_mask=None,
                token_type_ids=None,
                start_ids=None,
                end_ids=None,
                **kwargs):
        #  subjects_ids=None):
        outputs = self.bert(input_ids=input_ids,
                            attention_mask=attention_mask,
                            token_type_ids=token_type_ids)
        sequence_output = outputs[0]
        sequence_output = self.dropout(sequence_output)
        start_logits = self.start_fc(sequence_output)

        if start_ids is not None and self.training:
            if self.soft_label:
                batch_size = input_ids.size(0)
                seq_len = input_ids.size(1)
                label_logits = torch.FloatTensor(batch_size, seq_len,
                                                 self.num_labels)
                label_logits.zero_()
                label_logits = label_logits.to(input_ids.device)
                label_logits.scatter_(2, start_ids.unsqueeze(2), 1)
            else:
                label_logits = start_ids.unsqueeze(2).float()
        else:
            label_logits = F.softmax(start_logits, -1)
            if not self.soft_label:
                label_logits = torch.argmax(label_logits,
                                            -1).unsqueeze(2).float()
        end_logits = self.end_fc(sequence_output, label_logits)
        #  avg_logits = (start_logits + end_logits) / 2
        #  outputs = (
        #      #  avg_logits,
        #      start_logits,
        #      end_logits,
        #  ) + outputs[2:]
        outputs = (start_logits, end_logits)

        if start_ids is not None and end_ids is not None:
            assert self.loss_type in [
                'LabelSmoothingCrossEntropy', 'FocalLoss', 'CrossEntropyLoss'
            ]
            if self.loss_type == 'LabelSmoothingCrossEntropy':
                loss_fct = LabelSmoothingCrossEntropy()
            elif self.loss_type == 'FocalLoss':
                loss_fct = FocalLoss(gamma=self.focalloss_gamma)
                #  alpha=self.focalloss_alpha)
            elif self.loss_type == 'DiceLoss':
                loss_fct = DiceLoss(weight=self.diceloss_weight)
            else:
                loss_fct = CrossEntropyLoss()
            start_logits = start_logits.view(-1, self.num_labels)
            end_logits = end_logits.view(-1, self.num_labels)
            active_loss = attention_mask.view(-1) == 1
            active_start_logits = start_logits[active_loss]
            active_end_logits = end_logits[active_loss]

            active_start_labels = start_ids.view(-1)[active_loss]
            active_end_labels = end_ids.view(-1)[active_loss]

            start_loss = loss_fct(active_start_logits, active_start_labels)
            end_loss = loss_fct(active_end_logits, active_end_labels)
            total_loss = (start_loss + end_loss) / 2
            outputs = (total_loss, ) + outputs
            return outputs
        else:
            #  return (0.0, ) + outputs
            #  logger.warning(
            #      f"start_ids: {start_ids}, end_ids: {end_ids}"
            #  )
            #  return (torch.tensor(0.0).cuda(), ) + outputs
            return outputs

            #  return outputs


class BertSpanModel(TransformerModel):
    def __init__(
        self,
        model_name_or_path,
        num_labels,
        tokenizer=None,
        dropout_prob=0.1,
        #  loss_type='CrossEntropyLoss',
        loss_type='FocalLoss',
        #  loss_type='LabelSmoothingCrossEntropy',
        #  loss_type='DiceLoss',
        **kwargs):

        # for TransformerModel.load_from_config()
        self.num_labels = num_labels
        super(BertSpanModel, self).__init__(model_name_or_path,
                                            tokenizer=tokenizer,
                                            automodel_cls=BertSpanForNer)
        self.init_weights()

    def load_from_pretrained(self, model_path):
        self.transformer = BertSpanForNer.from_pretrained(
            model_path, num_labels=self.num_labels)

    def load_from_config(self):
        #  self.transformer = BertSpanForNer.from_config(self.config)
        setattr(self.config, 'num_labels', self.num_labels)
        self.transformer = BertSpanForNer(self.config)

    def forward(self,
                input_ids,
                attention_mask,
                token_type_ids,
                start_ids=None,
                end_ids=None,
                pseudo=None,
                tags=None):

        outputs = self.transformer(input_ids,
                                   attention_mask=attention_mask,
                                   token_type_ids=token_type_ids,
                                   start_ids=start_ids,
                                   end_ids=end_ids,
                                   tags=tags)
        if tags is not None:
            loss, start_logits, end_logits = outputs
            return loss, start_logits, end_logits
        else:
            start_logits, end_logits = outputs
            return start_logits, end_logits

    @classmethod
    def decode_ents_0_23_0(start_logits,
                           end_logits,
                           lens,
                           confidence,
                           overlap=False):
        num_tokens = lens[0]
        S = []

        starts = torch.argmax(start_logits, -1).cpu().numpy()[0]  #[1:-1]
        ends = torch.argmax(end_logits, -1).cpu().numpy()[0]  #[1:-1]

        # start_logits.shape: (1, sentence_length, num_labels)
        #  logger.debug(f"start_logits: {start_logits.shape}")
        #  start_max_probs = [
        #      f"{start_logits[0][i][x]:.4f}" for i, x in enumerate(starts)
        #  ]
        #  logger.debug(f"start_max_probs: {start_max_probs}")
        #  logger.debug(f"starts: {starts.shape} {starts}")
        starts = [
            x if start_logits[0][i][x] >= confidence else 0
            for i, x in enumerate(starts)
        ]
        ends = [
            x if end_logits[0][i][x] >= confidence else 0
            for i, x in enumerate(ends)
        ]

        #  starts = [starts[0]] + [ x if x != starts[i] else 0 for i, x in enumerate(starts[1:])]
        #  ends = [ends[0]] + [ x if x != ends[i] else 0 for i, x in enumerate(ends[1:])]
        def filter_process(starts):
            new_starts = []
            for i, x in enumerate(starts):
                is_dup = False
                if i < len(starts) - 1 and x == starts[i + 1]:
                    is_dup = True
                elif i > 0 and starts[i - 1] == x:
                    is_dup = True
                if is_dup:
                    new_starts.append(0)
                else:
                    new_starts.append(x)
            return new_starts

        #  starts = filter_process(starts)
        #  ends = filter_process(ends)

        #  starts = np.array([x for x in starts if x >= 0 and x < num_tokens])
        #  ends = np.array([x for x in ends if x >= 0 and x < num_tokens])
        starts = starts[:num_tokens]
        ends = ends[:num_tokens]

        #  logger.info(f"start_pred: {starts}")
        #  logger.info(f"end_pred: {ends}")
        #  for i, s_l in enumerate(starts):
        #      if s_l == 0:
        #          continue
        #      for j, e_l in enumerate(ends[i:]):
        #          if s_l == e_l:
        #              S.append((s_l, i, i + j))
        #              break
        #          if i + j < len(starts) - 1 and starts[i + j + 1] != 0:
        #              break
        #  for i in range(len(starts) - 1):
        last_j = -1
        for i in range(len(starts)):
            if i <= last_j:
                continue
            s_l = starts[i]
            if s_l == 0:
                continue
            for j, e_l in enumerate(ends[i:]):
                if s_l == e_l:
                    if not overlap:
                        #  if sum(starts[i + 1:i + j + 1]) == 0:
                        S.append((int(s_l), i, i + j))
                        last_j = j
                        i = j + 1
                        break
                    else:
                        if sum(starts[i + 1:i + j + 1]) != 0:
                            break
                        S.append((int(s_l), i, i + j))
                if i + j < len(starts) - 1 and starts[i + j + 1] != 0:
                    break
        #  S = [x for x in S if x[1] <= x[2]]

        #  for x in S:
        #      assert x[1] >= 0 and x[2] >= 0 and x[1] <= x[2], f"S: {S}"

        return S

    @classmethod
    def decode_ents(cls, start_probs, end_probs, batch_lens, confidence=0.0):
        """
        start_probs.shape: (batch_size, max_length, num_labels)
        end_probs.shape: (batch_size, max_length, num_labels)
        """
        #  logger.info(f"start_probs.shape: {start_probs.shape}")
        #  logger.info(f"end_probs.shape: {end_probs.shape}")

        #  start_probs = start_probs[:, 1:-1]
        #  end_probs = end_probs[:, 1:-1]

        #  logger.info(f"start_probs: {start_probs}")
        #  logger.info(f"end_probs: {end_probs}")
        """
        start_preds.shape: (batch_size, max_length -2)
        end_preds.shape: (batch_size, max_length -2)
        """
        start_preds = np.argmax(start_probs, -1)
        end_preds = np.argmax(end_probs, -1)
        #  logger.info(f"start_preds: {start_preds.shape} {start_preds[:3,:]}")
        #  logger.info(f"end_preds: {end_preds.shape} {end_preds[:3,:]}")

        if confidence > 0.0:
            for i, x in enumerate(start_preds):
                start_preds[i] = [
                    category if start_probs[i][j][category] > confidence else 0
                    for j, category in enumerate(start_preds[i])
                ]
            for i, x in enumerate(end_preds):
                end_preds[i] = [
                    category if end_probs[i][j][category] > confidence else 0
                    for j, category in enumerate(end_preds[i])
                ]

        final_predict_ents = []
        for start_pred, end_pred, text_len in zip(start_preds, end_preds,
                                                  batch_lens):
            start_pred = start_pred[:text_len]
            end_pred = end_pred[:text_len]
            predict_ents = defaultdict(list)
            last_j = -1
            for i, s_type in enumerate(start_pred):
                if s_type == 0:
                    continue
                if i <= last_j:
                    continue
                for j, e_type in enumerate(end_pred[i:]):
                    if s_type == e_type:
                        last_j = j
                        s = i
                        e = j + i
                        predict_ents[s_type].append((s, e))
                        break
                    if i + j < len(start_pred) - 1 and start_pred[i + j +
                                                                  1] != 0:
                        break
            final_predict_ents.append(predict_ents)
        #  logger.warning(f"final_predict_ents: {final_predict_ents}")

        return final_predict_ents


class ___BertSpanModel(TransformerModel):
    def __init__(
        self,
        model_name_or_path,
        num_labels,
        tokenizer=None,
        dropout_prob=0.1,
        #  loss_type='CrossEntropyLoss',
        loss_type='FocalLoss',
        #  loss_type='LabelSmoothingCrossEntropy',
        #  loss_type='DiceLoss',
        **kwargs):
        super(BertSpanModel, self).__init__(model_name_or_path,
                                            tokenizer=tokenizer,
                                            automodel_cls=AutoModel)
        config = self.config
        self.soft_label = False
        self.num_labels = num_labels
        self.loss_type = loss_type
        self.focalloss_gamma = 2.0
        self.focalloss_alpha = 0.25
        #  self.diceloss_weight = config.diceloss_weight

        #  self.bert = BertModel(config)
        self.dropout = nn.Dropout(config.hidden_dropout_prob)
        self.start_fc = PoolerStartLogits(config.hidden_size, self.num_labels)
        if self.soft_label:
            self.end_fc = PoolerEndLogits(config.hidden_size + self.num_labels,
                                          self.num_labels)
        else:
            self.end_fc = PoolerEndLogits(config.hidden_size + 1,
                                          self.num_labels)
        self.init_weights()

    def load_from_pretrained(self, model_path):
        #      automodel_cls = AutoModel
        #      super(BertSpanModel,
        #            self)._load_from_pretrained(model_path,
        #                                        automodel_cls=automodel_cls)
        self.transformer = AutoModel.from_pretrained(model_path)

    def load_from_config(self):
        self.transformer = AutoModel.from_config(self.config)

    def forward(self,
                input_ids,
                attention_mask,
                token_type_ids,
                start_ids=None,
                end_ids=None,
                pseudo=None,
                tags=None):
        #  subjects_ids=None):
        outputs = self.transformer(input_ids=input_ids,
                                   attention_mask=attention_mask,
                                   token_type_ids=token_type_ids)
        sequence_output = outputs[0]
        sequence_output = self.dropout(sequence_output)
        start_logits = self.start_fc(sequence_output)

        if start_ids is not None and self.training:
            if self.soft_label:
                batch_size = input_ids.size(0)
                seq_len = input_ids.size(1)
                label_logits = torch.FloatTensor(batch_size, seq_len,
                                                 self.num_labels)
                label_logits.zero_()
                label_logits = label_logits.to(input_ids.device)
                label_logits.scatter_(2, start_ids.unsqueeze(2), 1)
            else:
                label_logits = start_ids.unsqueeze(2).float()
        else:
            label_logits = F.softmax(start_logits, -1)
            if not self.soft_label:
                label_logits = torch.argmax(label_logits,
                                            -1).unsqueeze(2).float()
        end_logits = self.end_fc(sequence_output, label_logits)
        avg_logits = (start_logits + end_logits) / 2
        #  outputs = (
        #      avg_logits,
        #      start_logits,
        #      end_logits,
        #  ) + outputs[2:]
        outputs = (start_logits, end_logits)

        if start_ids is not None and end_ids is not None:
            assert self.loss_type in [
                'LabelSmoothingCrossEntropy', 'FocalLoss', 'CrossEntropyLoss',
                'DiceLoss'
            ]
            if self.loss_type == 'LabelSmoothingCrossEntropy':
                loss_fct = LabelSmoothingCrossEntropy()
            elif self.loss_type == 'FocalLoss':
                loss_fct = FocalLoss(gamma=self.focalloss_gamma)
                #  alpha=self.focalloss_alpha)
            elif self.loss_type == 'DiceLoss':
                loss_fct = DiceLoss()
            else:
                loss_fct = CrossEntropyLoss()
            start_logits = start_logits.view(-1, self.num_labels)
            end_logits = end_logits.view(-1, self.num_labels)
            active_loss = attention_mask.view(-1) == 1
            active_start_logits = start_logits[active_loss]
            active_end_logits = end_logits[active_loss]

            active_start_labels = start_ids.view(-1)[active_loss]
            active_end_labels = end_ids.view(-1)[active_loss]

            # FocalLoss + CrossEntropyLoss
            #  loss_fct_focal = FocalLoss(gamma=self.focalloss_gamma)
            #  loss_fct_ce = CrossEntropyLoss()
            #
            #  start_loss_0 = loss_fct_ce(active_start_logits,
            #                             active_start_labels)
            #  end_loss_0 = loss_fct_ce(active_end_logits, active_end_labels)
            #  start_loss_1 = loss_fct_focal(active_start_logits,
            #                                active_start_labels)
            #  end_loss_1 = loss_fct_focal(active_end_logits, active_end_labels)
            #
            #  start_loss = (start_loss_0 + start_loss_1) / 2
            #  end_loss = (end_loss_0 + end_loss_1) / 2
            #  #
            #  start_loss = loss_fct(active_start_logits, active_start_labels)
            #  end_loss = loss_fct(active_end_logits, active_end_labels)

            #  FocalLoss + CrossEntropyLoss + LabelSmoothingCrossEntropy
            #  loss_fct_focal = FocalLoss(gamma=self.focalloss_gamma)
            #  loss_fct_ce = CrossEntropyLoss()
            #  loss_fct_lsce = LabelSmoothingCrossEntropy()
            #
            #  start_loss_0 = loss_fct_ce(active_start_logits,
            #  active_start_labels)
            #  end_loss_0 = loss_fct_ce(active_end_logits, active_end_labels)
            #
            #  start_loss_1 = loss_fct_focal(active_start_logits,
            #  active_start_labels)
            #  end_loss_1 = loss_fct_focal(active_end_logits, active_end_labels)
            #
            #  start_loss_2 = loss_fct_lsce(active_start_logits,
            #  active_start_labels)
            #  end_loss_2 = loss_fct_lsce(active_end_logits, active_end_labels)
            #
            #  start_loss = (start_loss_0 + start_loss_1 + start_loss_2) / 3
            #  end_loss = (end_loss_0 + end_loss_1 + end_loss_2) / 3

            start_loss = loss_fct(active_start_logits, active_start_labels)
            end_loss = loss_fct(active_end_logits, active_end_labels)

            total_loss = (start_loss + end_loss) / 2
            outputs = (total_loss, ) + outputs
            return outputs
        else:
            #  return (0.0, ) + outputs
            #  logger.warning(
            #      f"start_ids: {start_ids}, end_ids: {end_ids}"
            #  )
            #  return (torch.tensor(0.0).cuda(), ) + outputs
            return outputs

    #  @classmethod
    #  def decode_ents(cls, start_logits, end_logits):
    #      """
    #      start_logits.shape: (batch_size, max_length, num_labels)
    #      end_logits.shape: (batch_size, max_length, num_labels)
    #      """
    #      #  logger.info(f"start_logits.shape: {start_logits.shape}")
    #      #  logger.info(f"end_logits.shape: {end_logits.shape}")
    #      start_logits = start_logits[:, 1:-1]
    #      end_logits = end_logits[:, 1:-1]
    #
    #      logger.info(f"start_logits: {start_logits}")
    #      logger.info(f"end_logits: {end_logits}")
    #      """
    #      start_preds.shape: (batch_size, max_length -2)
    #      end_preds.shape: (batch_size, max_length -2)
    #      """
    #      start_preds = np.argmax(start_logits, -1)
    #      end_preds = np.argmax(end_logits, -1)
    #      #  logger.info(f"start_preds: {start_preds.shape} {start_preds[:3,:]}")
    #      #  logger.info(f"end_preds: {end_preds.shape} {end_preds[:3,:]}")
    #
    #      confidence = 0.35
    #      for i, x in enumerate(start_preds):
    #          start_preds[i] = [
    #              category if start_logits[i][j][category] > confidence else 0
    #              for j, category in enumerate(start_preds[i])
    #          ]
    #      for i, x in enumerate(end_preds):
    #          end_preds[i] = [
    #              category if end_logits[i][j][category] > confidence else 0
    #              for j, category in enumerate(end_preds[i])
    #          ]
    #
    #      final_predict_ents = []
    #      for start_pred, end_pred in zip(start_preds, end_preds):
    #          predict_ents = defaultdict(list)
    #          for i, s_type in enumerate(start_pred):
    #              if s_type == 0:
    #                  continue
    #              for j, e_type in enumerate(end_pred[i:]):
    #                  if s_type == e_type:
    #                      s = i
    #                      e = j + i
    #                      predict_ents[s_type].append((s, e))
    #                      break
    #          final_predict_ents.append(predict_ents)
    #      #  logger.warning(f"final_predict_ents: {final_predict_ents}")
    #
    #      return final_predict_ents

    @classmethod
    def decode_ents(cls, start_probs, end_probs):
        """
        start_probs.shape: (batch_size, max_length, num_labels)
        end_probs.shape: (batch_size, max_length, num_labels)
        """
        #  logger.info(f"start_probs.shape: {start_probs.shape}")
        #  logger.info(f"end_probs.shape: {end_probs.shape}")
        start_probs = start_probs[:, 1:-1]
        end_probs = end_probs[:, 1:-1]

        #  logger.info(f"start_probs: {start_probs}")
        #  logger.info(f"end_probs: {end_probs}")
        """
        start_preds.shape: (batch_size, max_length -2)
        end_preds.shape: (batch_size, max_length -2)
        """
        start_preds = np.argmax(start_probs, -1)
        end_preds = np.argmax(end_probs, -1)
        #  logger.info(f"start_preds: {start_preds.shape} {start_preds[:3,:]}")
        #  logger.info(f"end_preds: {end_preds.shape} {end_preds[:3,:]}")

        confidence = 0.5
        for i, x in enumerate(start_preds):
            start_preds[i] = [
                category if start_probs[i][j][category] > confidence else 0
                for j, category in enumerate(start_preds[i])
            ]
        for i, x in enumerate(end_preds):
            end_preds[i] = [
                category if end_probs[i][j][category] > confidence else 0
                for j, category in enumerate(end_preds[i])
            ]

        final_predict_ents = []
        for start_pred, end_pred in zip(start_preds, end_preds):
            predict_ents = defaultdict(list)
            for i, s_type in enumerate(start_pred):
                if s_type == 0:
                    continue
                for j, e_type in enumerate(end_pred[i:]):
                    if s_type == e_type:
                        s = i
                        e = j + i
                        predict_ents[s_type].append((s, e))
                        break
            final_predict_ents.append(predict_ents)
        #  logger.warning(f"final_predict_ents: {final_predict_ents}")

        return final_predict_ents


#  class SpanModel(TransformerModel):
#      def __init__(self,
#                   model_name_or_path,
#                   num_labels,
#                   tokenizer=None,
#                   dropout_prob=0.1,
#                   loss_type='ce',
#                   **kwargs):
#          """
#          tag the subject and object corresponding to the predicate
#          :param loss_type: train loss type in ['ce', 'ls_ce', 'focal']
#          """
#          super(SpanModel, self).__init__(model_name_or_path,
#                                          tokenizer=tokenizer)
#
#          out_dims = self.config.hidden_size
#
#          mid_linear_dims = kwargs.pop('mid_linear_dims', 128)
#
#          self.num_labels = num_labels
#
#          self.mid_linear = nn.Sequential(nn.Linear(out_dims, mid_linear_dims),
#                                          nn.ReLU(), nn.Dropout(dropout_prob))
#
#          out_dims = mid_linear_dims
#
#          self.start_fc = nn.Linear(out_dims, num_labels)
#          self.end_fc = nn.Linear(out_dims, num_labels)
#
#          reduction = 'none'
#          if loss_type == 'ce':
#              self.criterion = nn.CrossEntropyLoss(reduction=reduction)
#          elif loss_type == 'ls_ce':
#              self.criterion = LabelSmoothingCrossEntropy(reduction=reduction)
#          else:
#              self.criterion = FocalLoss(reduction=reduction)
#
#          self.loss_weight = nn.Parameter(torch.FloatTensor(1),
#                                          requires_grad=True)
#          self.loss_weight.data.fill_(-0.2)
#
#          init_blocks = [self.mid_linear, self.start_fc, self.end_fc]
#
#          self._init_weights(init_blocks)
#
#      def forward(self,
#                  input_ids,
#                  attention_mask,
#                  token_type_ids,
#                  start_ids=None,
#                  end_ids=None,
#                  pseudo=None,
#                  tags=None):
#
#          #  logger.warning(
#          #      f"input_ids.shape: {input_ids.shape}, start_ids.shape: {start_ids.shape}, end_ids.shape: {end_ids.shape}"
#          #  )
#          bert_outputs = self.transformer(input_ids=input_ids,
#                                          attention_mask=attention_mask,
#                                          token_type_ids=token_type_ids)
#
#          seq_out = bert_outputs[0]
#
#          seq_out = self.mid_linear(seq_out)
#
#          start_logits = self.start_fc(seq_out)
#          end_logits = self.end_fc(seq_out)
#          #  logger.warning(
#          #      f"start_logits.shape: {start_logits.shape}, end_logits.shape: {end_logits.shape}"
#          #  )
#
#          out = (
#              start_logits,
#              end_logits,
#          )
#
#          if start_ids is not None and end_ids is not None:  #and self.training:
#
#              start_logits = start_logits.view(-1, self.num_labels)
#              end_logits = end_logits.view(-1, self.num_labels)
#
#              # 去掉 padding 部分的标签，计算真实 loss
#              active_loss = attention_mask.view(-1) == 1
#              active_start_logits = start_logits[active_loss]
#              active_end_logits = end_logits[active_loss]
#
#              active_start_labels = start_ids.view(-1)[active_loss]
#              active_end_labels = end_ids.view(-1)[active_loss]
#
#              if pseudo is not None:
#                  seq_len = input_ids.size(1)  # 512
#                  # (batch,)
#                  start_loss = self.criterion(start_logits,
#                                              start_ids.view(-1)).view(
#                                                  -1, seq_len).mean(dim=-1)
#                  end_loss = self.criterion(end_logits, end_ids.view(-1)).view(
#                      -1, seq_len).mean(dim=-1)
#
#                  # nums of pseudo data
#                  pseudo_nums = pseudo.sum().item()
#                  total_nums = input_ids.shape[0]
#
#                  # learning parameter
#                  rate = torch.sigmoid(self.loss_weight)
#                  if pseudo_nums == 0:
#                      start_loss = start_loss.mean()
#                      end_loss = end_loss.mean()
#                  else:
#                      if total_nums == pseudo_nums:
#                          start_loss = (rate * pseudo *
#                                        start_loss).sum() / pseudo_nums
#                          end_loss = (rate * pseudo *
#                                      end_loss).sum() / pseudo_nums
#                      else:
#                          start_loss = (rate*pseudo*start_loss).sum() / pseudo_nums \
#                                       + ((1 - rate) * (1 - pseudo) * start_loss).sum() / (total_nums - pseudo_nums)
#                          end_loss = (rate*pseudo*end_loss).sum() / pseudo_nums \
#                                       + ((1 - rate) * (1 - pseudo) * end_loss).sum() / (total_nums - pseudo_nums)
#              else:
#                  seq_len = input_ids.size(1)  # 512
#                  start_loss = self.criterion(start_logits,
#                                              start_ids.view(-1)).view(
#                                                  -1, seq_len).mean(dim=-1)
#                  end_loss = self.criterion(end_logits, end_ids.view(-1)).view(
#                      -1, seq_len).mean(dim=-1)
#                  start_loss = start_loss.mean()
#                  end_loss = end_loss.mean()
#                  #  start_loss = self.criterion(active_start_logits,
#                  #                              active_start_labels)
#                  #  end_loss = self.criterion(active_end_logits, active_end_labels)
#                  #  logger.info(
#                  #      f"start_loss.shape: {start_loss.shape}, active_start_logits.shape: {active_start_logits.shape}, active_start_labels.shape: {active_start_labels.shape}"
#                  #  )
#                  #  logger.info(
#                  #      f"end_loss.shape: {end_loss.shape}, active_end_logits.shape: {active_end_logits.shape}, active_end_labels.shape: {active_end_labels.shape}"
#                  #  )
#
#              loss = start_loss + end_loss
#
#              out = (loss, ) + out
#
#          return out
#
#      @classmethod
#      def decode_ents(cls, start_logits, end_logits):
#
#          #  logger.info(f"start_logits.shape: {start_logits.shape}")
#          #  logger.info(f"end_logits.shape: {end_logits.shape}")
#          start_logits = start_logits[:, 1:-1]
#          end_logits = end_logits[:, 1:-1]
#          start_preds = np.argmax(start_logits, -1)
#          end_preds = np.argmax(end_logits, -1)
#          #  logger.info(f"start_preds: {start_preds.shape} {start_preds[:3,:]}")
#          #  logger.info(f"end_preds: {end_preds.shape} {end_preds[:3,:]}")
#
#          final_predict_ents = []
#          for start_pred, end_pred in zip(start_preds, end_preds):
#              predict_ents = defaultdict(list)
#              for i, s_type in enumerate(start_pred):
#                  if s_type == 0:
#                      continue
#                  for j, e_type in enumerate(end_pred[i:]):
#                      if s_type == e_type:
#                          s = i
#                          e = j + i
#                          predict_ents[s_type].append((s, e))
#                          break
#              final_predict_ents.append(predict_ents)
#          #  logger.warning(f"final_predict_ents: {final_predict_ents}")
#
#          return final_predict_ents


# ------------------------------ TaskRunner ------------------------------
class NerRunner(TaskRunner):
    """
    任务专属模型定义
    """
    def __init__(self, task_args, ner_labels):
        super(NerRunner, self).__init__(**task_args.to_dict())
        logger.warning(f"ner_labels: {ner_labels}")
        self.ner_labels = ner_labels
        self.num_labels = len(self.ner_labels) + 1
        self.label2id = {x: i + 1 for i, x in enumerate(ner_labels)}
        self.id2label = {i + 1: x for i, x in enumerate(ner_labels)}
        self.type_weights = np.ones(self.num_labels)

        model_args = task_args.model_args
        #  self.model = SpanModel(
        self.model = BertSpanModel(
            model_name_or_path=model_args.model_name_or_path
            if model_args.model_name_or_path else model_args.checkpoint_path,
            num_labels=self.num_labels)

        self.all_ccm_list = []

    def forward(self, *args, **kwargs):
        kwargs = generate_method_kwargs_from_arguments(self.model.__class__,
                                                       "forward",
                                                       dict(**kwargs))
        return self.model(*args, **kwargs)

    def training_step(self, batch, batch_idx):
        outputs = self.forward(**batch)
        loss, start_logits, end_logits = outputs

        self.log('train_loss', loss, on_step=True)
        #  self.log('lr', self.hparams.lr, on_step=True)

        return OrderedDict({
            "loss": loss,
            'start_logits': start_logits,
            'end_logits': end_logits
        })

    def calculate_tp_fp_fn(self, golden_tags, pred_tags):
        """
        [(start, end), ...]
        """
        tp, fp, fn = 0, 0, 0
        for pred_tag in pred_tags:
            flag = 0
            for entity_gt in golden_tags:
                if pred_tag[0] == entity_gt[0] and pred_tag[1] == entity_gt[1]:
                    flag = 1
                    tp += 1
                    break
            if flag == 0:
                fp += 1

        fn = len(golden_tags) - tp

        return np.array([tp, fp, fn])

    def category_confuse_matrix(self, golden_tags, pred_tags):
        """
        {category_id: [(start, end), ...]}
        """
        ccm = {}
        for c_id, c in self.id2label.items():
            if c_id not in ccm:
                ccm[c_id] = np.array([0, 0, 0])
            ccm[c_id] += self.calculate_tp_fp_fn(golden_tags.get(c_id, []),
                                                 pred_tags.get(c_id, []))

        return ccm

    def validation_step(self, batch, batch_idx):
        outputs = self.forward(**batch)
        val_loss, start_logits, end_logits = outputs
        #  logger.warning(f"{val_loss}")
        self.log('val_loss', val_loss, on_step=True)

        start_probs = F.softmax(start_logits, -1)
        end_probs = F.softmax(end_logits, -1)
        #  logger.info(f"start_probs: {start_probs}")
        start_probs = start_probs.cpu().numpy()
        end_probs = end_probs.cpu().numpy()
        batch_lens = [am.sum() for am in batch['attention_mask']]
        batch_pred_tags = self.model.decode_ents(start_probs,
                                                 end_probs,
                                                 batch_lens=batch_lens)

        batch_tags = batch['tags']
        batch_char2tokens = batch['char2token']

        batch_golden_tags = []
        for tags, char2token in zip(batch_tags, batch_char2tokens):
            golden_tags = defaultdict(list)
            for tag in tags:
                c = tag['category']
                s = tag['start']
                m = tag['mention']
                e = s + len(m) - 1
                s = char2token[s]
                e = char2token[e]
                item = (s, e)
                golden_tags[self.label2id[c]].append(item)
            batch_golden_tags.append(golden_tags)

        all_ccm = []
        for golden_tags, pred_tags in zip(batch_golden_tags, batch_pred_tags):
            ccm = self.category_confuse_matrix(golden_tags, pred_tags)
            all_ccm.append(ccm)

        self.all_ccm_list.append(all_ccm)
        category_confuse_matrix = {
            c_id: np.array([0, 0, 0])
            for c_id in self.id2label.keys()
        }
        for all_ccm in self.all_ccm_list:
            for ccm in all_ccm:
                for c_id, x in ccm.items():
                    category_confuse_matrix[c_id] += x
        micro_metrics = np.zeros(3)
        category_p_r_f1 = {}
        for c_id, (tp, fp, fn) in category_confuse_matrix.items():
            p_r_f1 = get_p_r_f1(tp, fp, fn)
            category_p_r_f1[self.id2label[c_id]] = (p_r_f1, (tp, fp, fn))
            micro_metrics += p_r_f1 * self.type_weights[c_id - 1]
        micro_metrics /= self.num_labels
        val_precision, val_recall, val_f1 = micro_metrics
        self.log("val_precision", val_precision, on_step=True)
        self.log("val_recall", val_recall, on_step=True)
        self.log("val_f1", val_f1, on_step=True)

        return OrderedDict({
            'val_loss': val_loss,
            'val_precision': val_precision,
            'val_recall': val_recall,
            'val_f1': val_f1,
            'category_confuse_matrix': all_ccm
        })

    def show_val_results(self, category_p_r_f1):
        num_categories = self.num_labels - 1

        c_p_r_f1 = sorted(category_p_r_f1.items(),
                          key=lambda x: x[1][0][2],
                          reverse=True)
        max_key_len = max(max([len(c.encode('gbk')) + 1 for c, _ in c_p_r_f1]),
                          5)
        #  max_key_len = 30

        title = f"{' '*max_key_len}   P     R     F1   "
        title += " |   tp |   fp |   fn | true "
        title_len = len(title.encode('gbk'))

        #  n_english = len(re.findall('[^\u4e00-\u9fa5]', title))
        #  title_len = (len(title) - n_english) * 2 + n_english

        logger.info(f"{'=' * title_len}")
        logger.info(title)
        logger.info(f"{'-' * title_len}")

        total_tp = 0
        total_fp = 0
        total_fn = 0

        micro_metrics = np.zeros(3)
        for c, ((p, r, f1), (tp, fp, fn)) in c_p_r_f1:
            #  disp_key = c[:16]
            disp_key = c
            disp_key += ' ' * (max_key_len - len(disp_key.encode('gbk')))
            #  n_english = len(re.findall('[^\u4e00-\u9fa5]', disp_key))
            #  disp_key += ' ' * ((max_key_len -
            #                      (len(disp_key) - n_english) * 2 + n_english))
            info = f"{disp_key} | {p:.3f} {r:.3f} {f1:.3f}"
            info += f" | {tp:4d} | {fp:4d} | {fn:4d} | {tp+fn:4d}"
            logger.info(info)
            total_tp += tp
            total_fp += fp
            total_fn += fn

            c_id = self.label2id[c]
            p_r_f1 = np.array([p, r, f1])
            micro_metrics += p_r_f1 * self.type_weights[c_id - 1]

        logger.info(f"{'-' * title_len}")

        total_pred = total_tp + total_fp
        total_true = total_tp + total_fn

        micro_metrics /= num_categories
        micro_acc, micro_recall, micro_f1 = micro_metrics
        info = f"Micro{' '*(max_key_len - 6)} | {micro_acc:.3f} {micro_recall:.3f} {micro_f1:.3f}"  #" - loss: {results['loss']:.4f}"
        info += f" | {total_tp:4d} | {total_fp:4d} | {total_fn} | {total_true:4d}"
        logger.info(info)

        macro_acc = total_tp / total_pred if total_pred > 0.0 else 0.0
        macro_recall = total_tp / total_true if total_true > 0.0 else 0.0
        macro_f1 = 2 * macro_acc * macro_recall / (
            macro_acc + macro_recall) if (macro_acc +
                                          macro_recall) > 0.0 else 0.0
        info = f"Macro{' '*(max_key_len - 6)} | {macro_acc:.3f} {macro_recall:.3f} {macro_f1:.3f}"  #" - loss: {results['loss']:.4f}"
        logger.info(info)

    def validation_epoch_end(self, outputs):
        x = [out['val_loss'].shape for out in outputs]
        val_loss = sum([out["val_loss"] for out in outputs]) / len(outputs)
        self.log('val_loss', val_loss, on_epoch=True)

        category_confuse_matrix = {
            c_id: np.array([0, 0, 0])
            for c_id in self.id2label.keys()
        }
        for out in outputs:
            all_ccm = out['category_confuse_matrix']
            for ccm in all_ccm:
                for c_id, x in ccm.items():
                    category_confuse_matrix[c_id] += x

        micro_metrics = np.zeros(3)
        category_p_r_f1 = {}
        for c_id, (tp, fp, fn) in category_confuse_matrix.items():
            p_r_f1 = get_p_r_f1(tp, fp, fn)
            category_p_r_f1[self.id2label[c_id]] = (p_r_f1, (tp, fp, fn))
            micro_metrics += p_r_f1 * self.type_weights[c_id - 1]

        micro_metrics /= self.num_labels - 1

        val_precision, val_recall, val_f1 = micro_metrics
        #  logger.info(f"category_p_r_f1: {category_p_r_f1}")
        #  logger.info(f"micro metrics: {micro_metrics}")
        #  logger.info(
        #      f"val_precision: {val_precision:.4f}, val_recall: {val_recall:.4f}, val_f1: {val_f1:.4f}"
        #  )
        self.log("val_precision", val_precision, on_epoch=True)
        self.log("val_recall", val_recall, on_epoch=True)
        self.log("val_f1", val_f1, on_epoch=True)

        self.show_val_results(category_p_r_f1)

        #  logger.warning(
        #      f"trainer.callback_metrics: {self.trainer.callback_metrics}")
        eval_outputs = {
            'val_loss': val_loss,
            'val_precision': val_precision,
            'val_recall': val_recall,
            'val_f1': val_f1
        }
        self.save_best_model(eval_outputs)

    def test_step(self, batch, batch_idx):
        test_kwargs = {**batch}
        test_kwargs['start_ids'] = None
        test_kwargs['end_ids'] = None
        test_kwargs['tags'] = None
        #  start_logits, end_logits = self.forward(**batch)
        start_logits, end_logits = self.forward(**test_kwargs)

        start_probs = F.softmax(start_logits, -1)
        end_probs = F.softmax(end_logits, -1)
        start_probs = start_probs.cpu().numpy()
        end_probs = end_probs.cpu().numpy()

        batch_lens = [am.sum() for am in batch['attention_mask']]
        batch_pred_tags = self.model.decode_ents(start_probs,
                                                 end_probs,
                                                 batch_lens=batch_lens)

        batch_token2chars = batch['token2char']
        for pred_tags, token2char in zip(batch_pred_tags, batch_token2chars):
            for c, tags in pred_tags.items():
                pred_tags[c] = [(token2char[s], token2char[e + 1] - 1)
                                for s, e in tags]

        return OrderedDict({
            'preds': batch_pred_tags,
            'start_logits': start_logits,
            'end_logits': end_logits,
        })

    def test_epoch_end(self, outputs):

        test_preds = np.concatenate([out['preds'] for out in outputs])
        logger.info(f"test_preds: {test_preds.shape}")

        test_dataset = self.test_dataloader.dataloader.dataset
        seg_spans = test_dataset.seg_spans

        assert len(seg_spans) == test_preds.shape[
            0], f"len(seg_spans): {len(seg_spans)} == test_preds.shape[0]: {test_preds.shape[0]}"

        test_ents = defaultdict(list)
        for (guid, s_seg, e_seg), dict_ents in zip(seg_spans, test_preds):
            shift_ents = []
            for c, ents in dict_ents.items():
                ents = [(c, s + s_seg, e + s_seg) for s, e in ents
                        if s >= 0 and e >= 0]

                if ents:
                    shift_ents.extend(ents)

            test_ents[guid].append(shift_ents)

        # 合并去重
        for guid, ents in test_ents.items():
            ents = [x for x in ents if x]
            unique_ents = []
            ent_keys = []
            if len(ents) > 0:
                ents = np.concatenate(ents).tolist()
                for ent in ents:
                    key = str(ent)
                    if key not in ent_keys:
                        ent_keys.append(key)
                        unique_ents.append(ent)
            test_ents[guid] = unique_ents
        #  for guid, ents in test_ents.items():
        #      unique_ents = []
        #      ent_keys = []
        #      for ent in ents:
        #          key = str(ent)
        #          logger.info(f"key: {key}")
        #          if key not in ent_keys:
        #              ent_keys.append(key)
        #              unique_ents.append(ent)
        #          else:
        #              logger.warning(f"Duplicate {key}")
        #      test_ents[guid] = unique_ents

        #  assert len(test_ents) == len(
        #      test_dataset
        #  ), f"len(test_ents): {len(test_ents)}, len(test_dataset): {len(test_dataset)}"

        final_preds = test_ents

        #  final_start_logits = np.concatenate(
        #      [out['start_logits'] for out in outputs])
        #  logger.info(f"start_logits: {final_start_logits.shape}")
        #  final_start_logits = final_start_logits.tolist()
        #
        #  final_end_logits = np.concatenate(
        #      [out['end_logits'] for out in outputs])
        #  logger.info(f"end_logits: {final_end_logits.shape}")
        #  final_end_logits = final_end_logits.tolist()

        self.test_results = {
            'preds': final_preds,
            #  'start_logits': final_start_logits,
            #  'end_logits': final_end_logits
        }


# ------------------------------ Task ------------------------------
class NerTask(BaseTask):
    #  def __init__(self, *args, **kwargs):
    #      super(NerTask, self).__init__(*args, **kwargs)

    def __init__(self, args: Type[TaskArguments], data: Type[TaskData],
                 ner_labels: list):
        runner = NerRunner(args, ner_labels)
        super(NerTask, self).__init__(args, data, runner)

    def execute(self, *args, **kwargs):
        model_args = self.model_args
        data_args = self.data_args
        training_args = self.training_args
        remaining_args = self.remaining_args

        super(NerTask, self).execute(*args, **kwargs)

    def generate_submission(self):
        logger.warning(f"NerTask.generate_submission().")



#  -*- coding: utf-8 -*-
#  File: __init__.py

"""
Common utils.
These utils should be irrelevant to tensorflow.
"""

# https://github.com/celery/kombu/blob/7d13f9b95d0b50c94393b962e6def928511bfda6/kombu/__init__.py#L34-L36
STATICA_HACK = True
globals()['kcah_acitats'[::-1].upper()] = False
if STATICA_HACK:
    from .utils import *


__all__ = []


def _global_import(name):
    p = __import__(name, globals(), None, level=1)
    lst = p.__all__ if '__all__' in dir(p) else dir(p)
    for k in lst:
        if not k.startswith('__'):
            globals()[k] = p.__dict__[k]
            __all__.append(k)


_global_import('utils')

# Import no other submodules. they are supposed to be explicitly imported by users.
__all__.extend(['logger'])



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import json, copy
from collections import defaultdict
from pathlib import Path
from tqdm import tqdm
from loguru import logger
import numpy as np
import torch
from torch.utils.data import DataLoader, TensorDataset
from torch.utils.data import RandomSampler, SequentialSampler
from ...utils.multiprocesses import barrier_leader_process, barrier_member_processes, is_multi_processes
from ...utils import seg_generator


class InputExample(object):
    """A single training/test example for token classification."""
    def __init__(self, guid, text_a, spo_list, text_offset=0):
        self.guid = guid
        self.text_a = text_a
        self.spo_list = spo_list
        self.text_offset = text_offset

    def __repr__(self):
        return str(self.to_json_string())

    def to_dict(self):
        """Serializes this instance to a Python dictionary."""
        output = copy.deepcopy(self.__dict__)
        return output

    def to_json_string(self):
        """Serializes this instance to a JSON string."""
        return json.dumps(self.to_dict(), indent=2, sort_keys=True) + "\n"


class InputFeature(object):
    """A single set of features of data."""
    def __init__(self, input_ids, input_mask, input_len, segment_ids,
                 sub_labels, seed_sub, obj_labels):
        self.input_ids = input_ids
        self.input_mask = input_mask
        self.segment_ids = segment_ids
        self.input_len = input_len
        self.sub_labels = sub_labels
        self.seed_sub = seed_sub
        self.obj_labels = obj_labels

    def __repr__(self):
        return str(self.to_json_string())

    def __eq__(self, other):
        return self.input_ids == other.input_ids and \
            self.input_mask == other.input_mask and \
            self.segment_ids == other.segment_ids and \
            self.labels == other.labels and \
            self.sub_labels == other.sub_labels and \
            self.seed_sub == other.seed_sub and \
            self.obj_labels == other.obj_labels and \
            self.input_len == other.input_len

    def to_dict(self):
        """Serializes this instance to a Python dictionary."""
        output = copy.deepcopy(self.__dict__)
        return output

    def to_json_string(self):
        """Serializes this instance to a JSON string."""
        return json.dumps(self.to_dict(), indent=2, sort_keys=True) + "\n"


def encode_examples(examples, label2id, tokenizer, max_seq_length):

    num_labels = len(label2id)
    texts = [e.text_a for e in examples]

    all_tokens = [[tokenizer.cls_token] +
                  tokenizer.tokenize(text)[:max_seq_length - 2] +
                  [tokenizer.sep_token]
                  for text in tqdm(texts, desc="Tokenize")]
    all_input_lens = [len(tokens) for tokens in all_tokens]
    all_input_ids = [
        tokenizer.convert_tokens_to_ids(tokens) for tokens in all_tokens
    ]
    all_attention_mask = [[1] * len(input_ids) for input_ids in all_input_ids]
    all_token_type_ids = [[0] * len(input_ids) for input_ids in all_input_ids]

    all_padding_lens = [max_seq_length - n for n in all_input_lens]
    for i, (input_ids, attention_mask, token_type_ids,
            padding_length) in enumerate(
                zip(all_input_ids, all_attention_mask, all_token_type_ids,
                    all_padding_lens)):
        all_input_ids[i] = input_ids + [0] * padding_length
        all_attention_mask[i] = attention_mask + [0] * padding_length
        all_token_type_ids[i] = token_type_ids + [0] * padding_length

    def encode_spo(spo_list, num_tokens):
        """
        spo_list: [(s, p, o), ...]
            s: (sub_s, sub_e)
            p: predicate
            o: (obj_s, obj_e)

        """
        #  logger.warning(f"num_tokens: {num_tokens}")

        if spo_list:
            # 将spo三元组列表转换为词典{s: (p,o)}
            spoes = {s: (label2id[p], o) for s, p, o in spo}

            sub_labels = np.zeros((num_tokens, 2), dtype=int)
            for s in spoes:
                sub_labels[s[0] + 1, 0] = 1
                sub_labels[s[1] + 1, 1] = 1

            # 随机选一个subject
            start, end = np.array(list(spoes.keys())).T
            start = np.random.choice(start)
            end = np.random.choice(end[end >= start])
            seed_sub = (start, end)
            # 对应的object标签
            object_labels = np.zeros((num_tokens, num_labels, 2))
            for o in spoes.get(seed_sub, []):
                object_labels[o[0], o[2], 0] = 1
                object_labels[o[1], o[2], 1] = 1

        return sub_labels, seed_sub, obj_labels

    all_sub_labels = []
    all_seed_subs = []
    all_obj_labels = []

    all_spo_lists = [e.spo_list for e in examples]
    for input_ids, attention_mask, token_type_ids, spo_list in zip(
            all_input_ids, all_attention_mask, all_token_type_ids,
            all_spo_lists):
        num_tokens = len(input_ids)
        sub_labels, seed_sub, obj_labels = encode_spo(spo_list, num_tokens)

        assert len(input_ids) == max_seq_length
        assert len(attention_mask) == max_seq_length
        assert len(token_type_ids) == max_seq_length
        assert len(sub_labels) == max_seq_length
        assert len(obj_labels) == max_seq_length

        all_sub_labels.append(sub_labels)
        all_seed_subs.append(seed_sub)
        all_obj_labels.append(obj_labels)

    logger.debug(f"all_input_ids.shape: {np.array(all_input_ids).shape}")
    logger.debug(
        f"all_attention_mask.shape: {np.array(all_attention_mask).shape}")
    logger.debug(
        f"all_token_type_ids.shape: {np.array(all_token_type_ids).shape}")
    logger.debug(f"all_sub_labels.shape: {np.array(all_sub_labels).shape}")
    logger.debug(f"all_obj_labels.shape: {np.array(all_obj_labels).shape}")
    logger.debug(f"all_input_lens.shape: {np.array(all_input_lens).shape}")
    logger.debug(f"all_seed_subs.shape: {np.array(all_seed_subs).shape}")
    assert np.array(all_input_ids).shape[1] == max_seq_length
    assert np.array(all_attention_mask).shape[1] == max_seq_length
    assert np.array(all_token_type_ids).shape[1] == max_seq_length

    all_features = [
        InputFeature(input_ids=input_ids,
                     input_mask=attention_mask,
                     segment_ids=token_type_ids,
                     sub_labels=sub_labels,
                     seed_sub=seed_sub,
                     obj_labels=obj_labels,
                     input_len=input_len) for input_ids, attention_mask,
        token_type_ids, sub_labels, seed_sub, obj_labels, input_len in zip(
            all_input_ids, all_attention_mask, all_token_type_ids,
            all_sub_labels, all_seed_subs, all_obj_labels, all_input_lens)
    ]
    return (all_features, {
        'input_ids':
        torch.tensor(all_input_ids, dtype=torch.long),
        'attention_mask':
        torch.tensor(all_attention_mask, dtype=torch.long),
        'token_type_ids':
        torch.tensor(all_token_type_ids, dtype=torch.long),
        'sub_labels':
        torch.tensor(all_sub_labels, dtype=torch.long),
        "seed_sub":
        all_seed_subs,
        'obj_labels':
        torch.tensor(all_obj_labels, dtype=torch.long),
        "input_lens":
        torch.tensor(all_input_lens, dtype=torch.long),
    })


def examples_to_dataset(examples, label2id, tokenizer, max_seq_length):
    all_features, outputs = encode_examples(examples, label2id, tokenizer,
                                            max_seq_length)

    dataset = TensorDataset(
        outputs['input_ids'],
        outputs['attention_mask'],
        outputs['token_type_ids'],
        outputs['sub_labels'],
        outputs['seed_sub'],
        outputs['obj_labels'],
        outputs['input_lens'],
    )

    assert len(dataset) == len(all_features)

    return dataset, all_features



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import json
import os
from collections import Counter

import numpy as np
import torch
import torch.nn as nn
import torch.nn.functional as F
from loguru import logger
from torch.nn import BCELoss, BCEWithLogitsLoss, CrossEntropyLoss
from torch.utils.data import DataLoader, RandomSampler, SequentialSampler

from transformers import (AutoConfig, AutoModelForTokenClassification,
                          BertConfig, BertTokenizer, BertTokenizerFast)
#  from transformers.modeling_bert import BertModel, BertPreTrainedModel
from transformers.models.bert import BertModel, BertPreTrainedModel

from ...losses import DiceLoss, FocalLoss, LabelSmoothingCrossEntropy
from ...utils.multiprocesses import (barrier_leader_process,
                                     barrier_member_processes,
                                     is_multi_processes)
from ..models.linears import PoolerEndLogits, PoolerStartLogits
from ..trainer import Trainer, get_default_optimizer_parameters
from .dataset import batch_to_input_data

BertLayerNorm = torch.nn.LayerNorm


def get_active_logits(logits, num_labels):
    # Only keep active parts of the loss
    loss_sig = nn.Sigmoid()
    active_logits = logits.view(-1, num_labels * 2)
    active_logits = loss_sig(active_logits)
    active_logits = active_logits**2

    return active_logits


def batch_gather(data: torch.Tensor, index: torch.Tensor):
    index = index.unsqueeze(-1)
    index = index.expand(data.size()[0], index.size()[1], data.size()[2])
    #  logger.info(f"batch_gather() index.size(): {index.size()}")
    return torch.gather(data, 1, index)


def extract_subject_1(output, seed_subs):
    """根据seed_subs从output中取出subject的向量表征
    """
    output = output.detach().cpu()
    seed_subs = seed_subs.detach().cpu()

    #  logger.warning(f"seed_subs: {seed_subs}")
    #  logger.warning(
    #      f"output.shape: {output.shape}, seed_subs.shape: {seed_subs.shape}")

    start = batch_gather(output, seed_subs[:, :1])
    end = batch_gather(output, seed_subs[:, 1:])
    # so_res = merge_function([start,end])
    # subject = torch.cat([start, end], 2)
    #  output.cuda()
    #  seed_subs.cuda()
    #  start.cuda()
    #  end.cuda()
    return start, end


    # return so_res
class BertCasrelForRE(BertPreTrainedModel):
    def __init__(self, config):
        super(BertCasrelForRE, self).__init__(config)
        self.num_labels = config.num_labels
        self.loss_type = config.loss_type
        self.focalloss_gamma = config.focalloss_gamma
        self.focalloss_alpha = config.focalloss_alpha
        self.diceloss_weight = config.diceloss_weight

        self.bert = BertModel(config)
        self.dropout = nn.Dropout(config.hidden_dropout_prob)
        self.LayerNorm = BertLayerNorm(config.hidden_size, eps=1e-12)

        self.classifier = nn.Linear(config.hidden_size, 2)

        self.obj_classifier = nn.Linear(config.hidden_size,
                                        self.num_labels * 2)
        self.sub_pos_emb = nn.Embedding(256, config.hidden_size)
        self.relu = nn.ReLU()
        self.linear = nn.Linear(config.hidden_size, config.hidden_size)

        self.init_weights()

    def forward(
        self,
        input_ids,
        attention_mask=None,
        token_type_ids=None,
        sub_labels=None,
        seed_subs=None,
        obj_labels=None,
        do_forward_sub=True,
        do_forward_obj=True,
    ):
        batch_size = input_ids.size(0)
        outputs_bert = self.bert(input_ids=input_ids,
                                 attention_mask=attention_mask,
                                 token_type_ids=token_type_ids)
        #  logger.info(f"outputs_bert: {outputs_bert}")
        #  for i, x in enumerate(outputs_bert):
        #      logger.info(f"{i}: {x.shape}")
        #  logger.info(
        #      f"do_forward_sub: {do_forward_sub}, do_forward_obj: {do_forward_obj}"
        #  )
        sequence_output = outputs_bert[0]
        sequence_output = self.dropout(sequence_output)

        outputs = self.forward_sub(sequence_output, attention_mask, sub_labels,
                                   batch_size)
        if do_forward_obj:
            if sub_labels is not None and obj_labels is not None:
                loss = outputs[0]
            else:
                loss = None
            outputs_obj = self.forward_obj(outputs_bert, sequence_output,
                                           attention_mask, seed_subs,
                                           obj_labels, batch_size, loss)
            return outputs_obj  # (loss), scores, (hidden_states), (attentions)
        else:
            return outputs

    def forward_sub(self, sequence_output, attention_mask, sub_labels,
                    batch_size):

        logits = self.classifier(sequence_output)
        outputs = (logits,
                   )  # add hidden states and attention if they are here
        #  logger.warning(f"logits.shape: {logits.shape}")
        loss_fct = BCELoss(reduction='none')
        # loss_fct = BCEWithLogitsLoss(reduction='none')
        loss_sig = nn.Sigmoid()
        # Only keep active parts of the loss
        #  active_logits = logits.view(-1, self.num_labels)
        active_logits = logits.view(-1, 2)
        active_logits = loss_sig(active_logits)
        active_logits = active_logits**2
        #  logger.warning(f"active_logits.shape: {active_logits.shape}")
        if sub_labels is not None:
            #  active_labels = labels.view(-1, self.num_labels).float()
            active_labels = sub_labels.view(-1, 2).float()
            #  logger.warning(
            #      f"active_logits.shape: {active_logits.shape}, active_labels.shape: {active_labels.shape}"
            #  )
            loss = loss_fct(active_logits, active_labels)
            # loss = loss.view(-1,sequence_output.size()[1],2)
            loss = loss.view(batch_size, -1, 2)
            loss = torch.mean(loss, 2)
            loss = torch.sum(attention_mask * loss) / torch.sum(attention_mask)
            outputs = (loss, ) + outputs
        else:
            outputs = (torch.tensor(0.0), active_logits)
            #  outputs = (torch.tensor(0.0), logits)

        return outputs
        #  logits = self.classifier(sequence_output)
        #  outputs = (logits,
        #             )  # add hidden states and attention if they are here
        #  loss_fct = BCELoss(reduction='none')
        #  # loss_fct = BCEWithLogitsLoss(reduction='none')
        #  loss_sig = nn.Sigmoid()
        #  # Only keep active parts of the loss
        #  active_logits = logits.view(-1, self.num_labels)
        #  active_logits = loss_sig(active_logits)
        #  active_logits = active_logits**2
        #  if sub_labels is not None:
        #      active_labels = sub_labels.view(-1, self.num_labels).float()
        #      loss = loss_fct(active_logits, active_labels)
        #      # loss = loss.view(-1,sequence_output.size()[1],2)
        #      loss = loss.view(batch_size, -1, 2)
        #      loss = torch.mean(loss, 2)
        #      loss = torch.sum(attention_mask * loss) / torch.sum(attention_mask)
        #      outputs = (loss, ) + outputs
        #  else:
        #      outputs = active_logits
        #      outputs = (torch.tensor(0.0).cuda(), ) + active_logits
        #
        #  return outputs

    def forward_obj(self,
                    outputs_bert,
                    sequence_output,
                    attention_mask,
                    seed_subs,
                    obj_labels,
                    batch_size,
                    loss=None):

        #  logger.info(f"1-obj_labels: {obj_labels}")
        hidden_states = outputs_bert[2][-2]
        hidden_states_1 = outputs_bert[2][-3]
        # hidden_states = self.dropout(hidden_states)
        loss_obj = BCELoss(reduction='none')
        loss_sig = nn.Sigmoid()
        sub_pos_start = self.sub_pos_emb(seed_subs[:, :1]).cuda()
        sub_pos_end = self.sub_pos_emb(seed_subs[:, 1:]).cuda()

        #  logger.info(f"1.0-seed_subs: {seed_subs}")
        #  logger.info(f"1.0-obj_labels: {obj_labels}")

        subject_start_last, subject_end_last = extract_subject_1(
            sequence_output, seed_subs)

        #  logger.info(f"1.0.1-seed_subs: {seed_subs}")
        #  logger.info(f"1.0.1-obj_labels: {obj_labels}")

        subject_start_1, subject_end_1 = extract_subject_1(
            hidden_states_1, seed_subs)

        #  logger.info(f"1.0.2-obj_labels: {obj_labels}")

        subject_start, subject_end = extract_subject_1(hidden_states,
                                                       seed_subs)
        #  logger.info(f"1.1-obj_labels: {obj_labels}")
        # subject = extract_subject(hidden_states, seed_subs)
        # subject_start = subject_start.cuda()
        # subject_end = subject_end.cuda())

        subject_start_last = subject_start_last.cuda()
        subject_end_last = subject_end_last.cuda()
        subject_start_1 = subject_start_1.cuda()
        subject_end_1 = subject_end_1.cuda()
        subject_start = subject_start.cuda()
        subject_end = subject_end.cuda()

        #  logger.info(
        #      f"sub_pos_start.shape: {sub_pos_start.shape}, sub_pos_end.shape: {sub_pos_end.shape}, subject_start.shape: {subject_start.shape}, subject_end.shape: {subject_end.shape}"
        #  )
        #  logger.info(f"subject_start_last.shape: {subject_start_last.shape}")
        #  logger.info(
        #      f"subject_start_1.shape: {subject_start_1.shape}, subject_end_1.shape: {subject_end_1.shape}"
        #  )
        subject = (sub_pos_start + subject_start + sub_pos_end + subject_end +
                   subject_start_last + subject_start_1 + subject_end_1 +
                   subject_end_1).cuda()
        # subject = extract_subject(sequence_output, seed_subs).cuda()
        batch_token_ids_obj = torch.add(hidden_states, subject)
        batch_token_ids_obj = self.LayerNorm(batch_token_ids_obj)
        batch_token_ids_obj = self.dropout(batch_token_ids_obj)

        #  logger.info(f"1.2-obj_labels: {obj_labels}")

        # batch_token_ids_obj = F.dropout(batch_token_ids_obj,p=0.5)
        batch_token_ids_obj = self.relu(self.linear(batch_token_ids_obj))
        batch_token_ids_obj = self.dropout(batch_token_ids_obj)
        # batch_token_ids_obj = F.dropout(batch_token_ids_obj,p=0.4)
        obj_logits = self.obj_classifier(batch_token_ids_obj)
        #  logger.warning(f"obj_labels.shape: {obj_labels.shape}")
        #  logger.info(f"obj_logits.shape: {obj_logits.shape}")

        obj_logits = obj_logits.view(-1, 2)
        obj_logits = loss_sig(obj_logits)
        obj_logits = obj_logits**4
        obj_outputs = (obj_logits, )

        #  active_logits = logits.view(-1, 2)
        #  active_logits = loss_sig(active_logits)
        #  active_logits = active_logits**2
        #  logger.info(f"2-obj_labels: {obj_labels}")
        if obj_labels is not None:
            #  obj_logits = obj_logits.view(-1,2)
            #  obj_logits = loss_sig(obj_logits)
            #  obj_logits = obj_logits**4
            #  obj_outputs = (obj_logits, )

            #  logger.warning(f"obj_labels: {obj_labels}")
            #  logger.info(
            #      f"obj_labels: {obj_labels.shape}, obj_logits: {obj_logits.shape}"
            #  )
            # obj_loss = loss_obj(obj_logits.view(-1, hidden_states.size()[1], self.obj_labels // 2, 2), obj_labels.float())
            obj_loss = loss_obj(
                #  obj_logits.view(batch_size, -1, self.num_labels // 2, 2),
                obj_logits.view(batch_size, -1, self.num_labels, 2),
                obj_labels.float())

            #  logger.warning(f"obj_loss: {obj_loss.shape}")
            #  logger.warning(f"obj_loss: {obj_loss[0]}")
            #  logger.warning(f"obj_loss: {obj_loss.shape}, {obj_loss}")
            obj_loss = torch.sum(torch.mean(obj_loss, 3), 2)
            #  logger.info(f"obj_loss: {obj_loss.shape}, {obj_loss}")
            obj_loss = torch.sum(
                obj_loss * attention_mask) / torch.sum(attention_mask)
            #  logger.info(f"obj_loss: {obj_loss.shape}, {obj_loss}")
            s_o_loss = torch.add(obj_loss, loss)
            #  logger.info(f"s_o_loss: {s_o_loss.shape}, {s_o_loss}")
            outputs_obj = (s_o_loss, ) + obj_outputs
        else:
            # outputs_obj = obj_logits.view(-1,hidden_states.size()[1],self.obj_labels // 2 ,2)
            outputs_obj = obj_logits.view(batch_size, -1, self.num_labels, 2)
            #  outputs_obj = obj_logits.view(batch_size, -1, self.num_labels // 2,
            #                                2)

        return outputs_obj


def extract_spoes(args, model, text, confidence=0.5):
    """抽取输入text所包含的三元组
    """
    tokens = tokenizer_k.tokenize(text, max_length=maxlen)
    # tokens_1 = tokenizer.tokenize(text,max_length=maxlen)
    mapping = tokenizer_k.rematch(text, tokens)
    token_ids, segment_ids = tokenizer_k.encode(text, max_length=maxlen)
    # token_ids_1 = tokenizer.encode_plus(text,max_length=maxlen)['input_ids']

    #  token_ids_1 = token_ids
    #  segment_ids_1 = segment_ids
    token_ids = torch.tensor([token_ids]).cuda()
    segment_ids = torch.tensor([segment_ids]).cuda()

    subject_preds = model(input_ids=token_ids,
                          token_type_ids=segment_ids,
                          do_forward_obj=False)

    subjects = extract_subjects_from_logits(args,
                                            subject_preds,
                                            confidence=confidence)
    spoes = extract_predicate_objects_from_logits(args,
                                                  model,
                                                  subjects,
                                                  mapping,
                                                  confidence=confidence)

    return spoes


def extract_subjects_from_logits(
    args,
    subject_preds,
    confidence=0.5,
):
    # 抽取subject
    subject_preds = subject_preds.view(1, -1, 2)
    subject_preds = subject_preds.detach().cpu().numpy()
    #  logger.info(f"subject_preds.shape: {subject_preds.shape}")

    start = np.where(subject_preds[0, :, 0] > confidence)[0]
    end = np.where(subject_preds[0, :, 1] > confidence)[0]
    subjects = []
    for i in start:
        j = end[end >= i]
        if len(j) > 0:
            j = j[0]
            subjects.append((i, j))
    return subjects


def extract_predicate_objects_from_logits(args,
                                          model,
                                          subjects,
                                          input_ids,
                                          attention_mask,
                                          token_type_ids,
                                          token2char,
                                          lens,
                                          confidence=0.5):

    #  logger.info(f"input_ids: {input_ids}")

    input_ids = input_ids.detach().cpu().numpy().tolist()
    attention_mask = attention_mask.detach().cpu().numpy().tolist()
    token_type_ids = token_type_ids.detach().cpu().numpy().tolist()

    sub_input_ids = np.repeat(input_ids, len(subjects), 0)
    #  sub_input_ids = np.repeat([input_ids], len(subjects), 0)
    sub_attention_mask = np.repeat([attention_mask], len(subjects), 0)
    sub_token_type_ids = np.repeat(token_type_ids, len(subjects), 0)
    #  sub_token_type_ids = np.repeat([token_type_ids], len(subjects), 0)

    sub_input_ids = torch.tensor(sub_input_ids).cuda()
    sub_attention_mask = torch.tensor(sub_attention_mask).cuda()
    sub_token_type_ids = torch.tensor(sub_token_type_ids).cuda()

    subjects = torch.tensor(subjects).cuda()
    # 传入subject，抽取object和predicate
    #  logger.info(
    #      f"sub_input_ids.shape: {sub_input_ids.shape}, sub_token_type_ids.shape: {sub_token_type_ids.shape}"
    #  )
    #  #  sub_attention_mask.shape: {sub_attention_mask.shape}
    #  logger.info(f"sub_input_ids: {sub_input_ids}")
    #  logger.info(f"sub_token_type_ids: {sub_token_type_ids}")
    #  logger.info(f"subjects: {subjects}")
    object_preds = model(input_ids=sub_input_ids,
                         attention_mask=sub_attention_mask,
                         token_type_ids=sub_token_type_ids,
                         seed_subs=subjects,
                         do_forward_sub=True,
                         do_forward_obj=True)
    #  logger.warning(f"object_preds:{object_preds.shape}, {object_preds}")

    text_len = lens[0].item()
    spoes = []
    object_preds = object_preds.detach().cpu().numpy()
    for subject, object_pred in zip(subjects, object_preds):
        start = np.where(object_pred[:, :, 0] > confidence)
        end = np.where(object_pred[:, :, 1] > confidence)
        #  logger.info(f"text_len: {text_len}, start: {start}, end: {end}")
        #  start = np.array([x for x in start if x >= 0 and x < text_len])
        #  end = np.array([x for x in end if x >= 0 and x < text_len])

        last_end = -1
        for _start, predicate1 in zip(*start):
            if _start <= last_end:
                continue
            if _start >= text_len:
                continue
            found = False
            for _end, predicate2 in zip(*end):
                if _end >= text_len:
                    continue
                if _start <= _end and predicate1 == predicate2:
                    s_subject, e_subject = subject
                    #  s_subject = token2char[s_subject][0].item()
                    #  e_subject = token2char[e_subject][-1].item()
                    #  _start = token2char[_start][0].item()
                    #  _end = token2char[_end][-1].item()
                    if s_subject >= 0 and e_subject >= 0 and _start >= 0 and _end >= 0:
                        #  logger.info(
                        #      f"spo: ({s_subject}:{e_subject}), {predicate1}, ({_start}:{_end})"
                        #  )
                        found = True
                        spoes.append(((s_subject, e_subject), predicate1,
                                      (_start, _end)))
                        last_end = _end

                    #  if len(mapping[subject[0]]) > 0 and len(
                    #          mapping[subject[1]]) > 0 and len(
                    #              mapping[_start]) > 0 and len(
                    #                  mapping[_end]) > 0:
                    #      spoes.append(((mapping[subject[0]][0],
                    #                     mapping[subject[1]][-1]), predicate1,
                    #                    (mapping[_start][0], mapping[_end][-1])))
                    break
            #  if found:
            #      break
    return spoes
    #  return [(text[s[0]:s[1] + 1], id2predicate[p], text[o[0]:o[1] + 1])
    #          for s, p, o, in spoes]


def combine_spoes(spoes):
    """合并SPO成官方格式
    """
    new_spoes = {}
    for s, p, o in spoes:
        p1, predicate, p2 = p.split('_')
        if (s, p1) in new_spoes:
            new_spoes[(s, p1)][p2] = o
        else:
            new_spoes[(s, p1)] = {p2: o}

    return [(k[0], k[1], v) for k, v in new_spoes.items()]


class SPO(tuple):
    """用来存三元组的类
    表现跟tuple基本一致，只是重写了 __hash__ 和 __eq__ 方法，
    使得在判断两个三元组是否等价时容错性更好。
    """
    def __init__(self, spo, tokenizer_k):
        self.spox = (
            tuple(tokenizer_k.tokenize(spo[0])),
            spo[1],
            tuple(
                sorted([(k, tuple(tokenizer_k.tokenize(v)))
                        for k, v in spo[2].items()])),
        )

    def __hash__(self):
        return self.spox.__hash__()

    def __eq__(self, spo):
        return self.spox == spo.spox


def evaluate(data, tokenizer_k):
    """评估函数，计算f1、precision、recall
    """
    X, Y, Z = 1e-10, 1e-10, 1e-10
    f = open('dev_pred_2.json', 'w', encoding='utf-8')
    pbar = tqdm()
    for d in data:

        R = combine_spoes(extract_spoes(d['text']))
        T = combine_spoes(d['spo_list'])
        R = set([SPO(spo, tokenizer_k) for spo in R])
        T = set([SPO(spo, tokenizer_k) for spo in T])
        X += len(R & T)
        Y += len(R)
        Z += len(T)
        f1, precision, recall = 2 * X / (Y + Z), X / Y, X / Z
        pbar.update()
        pbar.set_description('f1: %.5f, precision: %.5f, recall: %.5f' %
                             (f1, precision, recall))
        s = json.dumps(
            {
                'text': d['text'],
                'spo_list': list(T),
                'spo_list_pred': list(R),
                'new': list(R - T),
                'lack': list(T - R),
            },
            ensure_ascii=False,
            indent=4)
        f.write(s + '\n')
    pbar.close()
    f.close()
    return f1, precision, recall


class SpanEntityScore(object):
    def __init__(self, id2label):
        self.id2label = id2label
        self.reset()

    def reset(self):
        self.origins = []
        self.founds = []
        self.rights = []

    def compute(self, origin, found, right):
        recall = 0 if origin == 0 else (right / origin)
        precision = 0 if found == 0 else (right / found)
        f1 = 0. if recall + precision == 0 else (2 * precision *
                                                 recall) / (precision + recall)
        return recall, precision, f1

    def result(self):
        class_info = {}
        origin_counter = Counter([self.id2label[x[0]] for x in self.origins])
        found_counter = Counter([self.id2label[x[0]] for x in self.founds])
        right_counter = Counter([self.id2label[x[0]] for x in self.rights])
        for type_, count in origin_counter.items():
            origin = count
            found = found_counter.get(type_, 0)
            right = right_counter.get(type_, 0)
            recall, precision, f1 = self.compute(origin, found, right)
            class_info[type_] = {
                "acc": round(precision, 4),
                'recall': round(recall, 4),
                'f1': round(f1, 4)
            }
        origin = len(self.origins)
        found = len(self.founds)
        right = len(self.rights)
        recall, precision, f1 = self.compute(origin, found, right)
        return {'acc': precision, 'recall': recall, 'f1': f1}, class_info

    def update(self, true_subject, pred_subject):
        self.origins.extend(true_subject)
        self.founds.extend(pred_subject)
        self.rights.extend([
            pre_entity for pre_entity in pred_subject
            if pre_entity in true_subject
        ])


MODEL_CLASSES = {
    #  'bert': (BertConfig, BertCasrelForRE, CNerTokenizer),
    'bert': (BertConfig, BertCasrelForRE, BertTokenizer),
    #  'bert': (BertConfig, BertSpanForNer, BertTokenizerFast),
}


def load_pretrained_tokenizer(args):
    #  config_class, model_class, tokenizer_class = MODEL_CLASSES[args.model_type]
    #  tokenizer = tokenizer_class.from_pretrained(
    #      args.model_path,
    #      do_lower_case=args.do_lower_case,
    #      is_english=args.is_english,
    #      cache_dir=args.cache_dir if args.cache_dir else None,
    #  )

    from ..token_utils import HFTokenizer
    tokenizer = HFTokenizer(os.path.join(args.model_path, 'vocab.txt'),
                            lowercase=args.do_lower_case)
    #  from tokenizers import BertWordPieceTokenizer
    #  tokenizer = BertWordPieceTokenizer(
    #      os.path.join(args.model_path, "vocab.txt"))
    return tokenizer


def load_pretrained_model(args):
    # make sure only the first process in distributed training
    # will download model & vocab
    barrier_member_processes(args)

    config_class, model_class, tokenizer_class = MODEL_CLASSES[args.model_type]
    config = config_class.from_pretrained(
        args.model_path,
        num_labels=args.num_labels,
        loss_type=args.loss_type,
        cache_dir=args.cache_dir if args.cache_dir else None,
        output_hidden_states=True,
    )
    setattr(config, 'label2id', args.label2id)
    setattr(config, 'id2label', args.id2label)
    setattr(config, 'loss_type', args.loss_type)
    setattr(config, 'focalloss_gamma', args.focalloss_gamma)
    setattr(config, 'focalloss_alpha', args.focalloss_alpha)
    setattr(config, 'diceloss_weight', args.diceloss_weight)
    logger.info(f"model_path: {args.model_path}")
    logger.info(f"config:{config}")
    model = model_class.from_pretrained(
        args.model_path,
        from_tf=bool(".ckpt" in args.model_path),
        config=config,
        cache_dir=args.cache_dir if args.cache_dir else None,
    )

    # make sure only the first process in distributed training
    # will download model & vocab
    barrier_leader_process(args)

    return model


def collate_fn(batch):
    """
    batch should be a list of (sequence, target, length) tuples...
    Returns a padded tensor of sequences sorted from longest to shortest,
    """
    all_input_ids, all_input_mask, all_segment_ids, all_sub_labels, all_seed_subs, all_obj_labels, all_lens, all_token_offsets, all_texts, all_spo_lists = batch_to_input_data(
        batch)
    #  all_input_ids, all_input_mask, all_segment_ids, all_sub_labels, all_seed_subs, all_obj_labels, all_lens, all_token2char, all_token_offsets = map(
    #      torch.stack, zip(*batch))

    max_len = max(all_lens).item()
    all_input_ids = all_input_ids[:, :max_len]
    all_input_mask = all_input_mask[:, :max_len]
    all_segment_ids = all_segment_ids[:, :max_len]
    all_sub_labels = all_sub_labels[:, :max_len]
    all_obj_labels = all_obj_labels[:, :max_len]

    return all_input_ids, all_input_mask, all_segment_ids, all_sub_labels, all_seed_subs, all_obj_labels, all_lens, all_token_offsets, all_texts, all_spo_lists


def extract_entity_from_logits(args,
                               preds,
                               lens,
                               confidence=0.5,
                               enable_nested_entities=False):
    num_labels = int(preds.shape[-1] / 2)
    text_len = lens[0]

    #  logger.info(f"preds: (shape: {preds.shape})")  # |{preds})")
    preds = preds.view(1, -1, num_labels * 2)
    preds = preds.detach().cpu().numpy()

    starts = []
    ends = []
    for i in range(num_labels):
        start = np.where(preds[0, :, 2 * i] > confidence)[0]
        end = np.where(preds[0, :, 2 * i + 1] > confidence)[0]

        start = np.array([x for x in start if x >= 0 and x < text_len])
        end = np.array([x for x in end if x >= 0 and x < text_len])

        starts.append(start)
        ends.append(end)
    #  logger.info(f"starts: {starts}")
    #  logger.info(f"ends: {ends}")
    entities = []
    for n in range(num_labels):
        start = starts[n]
        end = ends[n]
        for idx, i in enumerate(start):
            jj = end[end >= i]
            for j in jj:
                if idx < len(start) - 1 and j >= start[idx + 1]:
                    break
                entities.append((n + 1, i - 1, j - 1))
                if not enable_nested_entities:
                    break

    entities = sorted(entities, key=lambda x: x[1])
    return entities


def load_model(args):
    model = load_pretrained_model(args)
    model.to(args.device)
    return model


def build_default_model(args):
    """
    自定义模型
    规格要求返回模型(model)、优化器(optimizer)、调度器(scheduler)三元组。
    """

    # -------- model --------
    model = load_pretrained_model(args)
    model.to(args.device)

    # -------- optimizer --------
    from transformers.optimization import AdamW
    optimizer_parameters = get_default_optimizer_parameters(
        model, args.weight_decay)
    optimizer = AdamW(optimizer_parameters,
                      lr=args.learning_rate,
                      eps=args.adam_epsilon,
                      correct_bias=False)

    # -------- scheduler --------
    from transformers.optimization import get_linear_schedule_with_warmup

    scheduler = get_linear_schedule_with_warmup(
        optimizer,
        num_warmup_steps=args.total_steps * args.warmup_rate,
        num_training_steps=args.total_steps)

    return model, optimizer, scheduler


def init_labels(args, labels):

    args.ner_labels = labels
    args.id2label = {i: label for i, label in enumerate(args.ner_labels)}
    args.label2id = {label: i for i, label in enumerate(args.ner_labels)}
    args.num_labels = len(args.label2id)

    logger.info(f"args.label2id: {args.label2id}")
    logger.info(f"args.id2label: {args.id2label}")
    logger.info(f"args.num_labels: {args.num_labels}")


def evaluate_spo_list(pred, target):
    """
    []((sub_start, sub_mention), predicate, (obj_start, obj_mention)), ...]
    """

    R = set(pred)
    T = set(target)
    X = len(R & T)
    Y = len(R)
    Z = len(T)

    return X, Y, Z


class SpoTrainer(Trainer):
    def __init__(self, args, ner_labels, build_model=None, tokenizer=None):
        super(SpoTrainer, self).__init__(args)
        init_labels(args, ner_labels)
        if tokenizer:
            self.tokenizer = tokenizer
        else:
            self.tokenizer = load_pretrained_tokenizer(args)

        #  from bert4keras.tokenizers import Tokenizer
        #  self.tokenizer_k = Tokenizer(os.path.join(args.model_path,
        #                                            'vocab.txt'),
        #                               do_lower_case=True)

        if build_model is None:
            self.build_model = build_default_model
        else:
            self.build_model = build_model

        self.label2id = args.label2id
        self.collate_fn = collate_fn

    def examples_to_dataset(self, examples, max_seq_length):
        from .dataset import examples_to_dataset
        return examples_to_dataset(examples, self.label2id, self.tokenizer,
                                   max_seq_length)

    def encode_examples(self, examples, max_seq_length):
        from .dataset import encode_examples
        return encode_examples(examples, self.label2id, self.tokenizer,
                               max_seq_length)

    def batch_to_inputs(self, args, batch, known_labels=True):
        #  inputs = {
        #      "input_ids": torch.stack([e.input_ids for e in batch]),
        #      "attention_mask": torch.stack([e.input_mask for e in batch]),
        #      "sub_labels": torch.stack([e.sub_labels for e in batch]),
        #      "seed_subs": torch.stack([e.seed_subs for e in batch]),
        #      "obj_labels": torch.stack([e.obj_labels for e in batch]),
        #  }
        #  if args.model_type != "distilbert":
        #      # XLM and RoBERTa don"t use segment_ids
        #      inputs["token_type_ids"] = (torch.stack([
        #          e.segment_ids for e in batch
        #      ]) if args.model_type in ["bert", "xlnet"] else None)
        inputs = {
            "input_ids": batch[0],
            "attention_mask": batch[1],
            "sub_labels": batch[3],
            "seed_subs": batch[4],
            "obj_labels": batch[5],
        }
        if args.model_type != "distilbert":
            # XLM and RoBERTa don"t use segment_ids
            inputs["token_type_ids"] = (
                batch[2] if args.model_type in ["bert", "xlnet"] else None)

        return inputs

    #  def generate_dataloader(self, args, dataset, batch_size, keep_order=True):
    #
    #      Sampler = SequentialSampler if keep_order else RandomSampler
    #      sampler = DistributedSampler(dataset) if is_multi_processes(
    #          args) else Sampler(dataset)
    #      dataloader = DataLoader(dataset,
    #                              sampler=sampler,
    #                              batch_size=batch_size,
    #                              collate_fn=collate_fn)
    #      return dataloader

    def on_eval_start(self, args, eval_dataset):
        self.eval_results = []
        self.metric = SpanEntityScore(args.id2label)
        self.X, self.Y, self.Z = 1e-10, 1e-10, 1e-10

    #  def on_eval_step(self, args, model, step, batch, batch_features):
    def on_eval_step(self, args, model, step, batch):

        all_input_ids, all_input_mask, all_segment_ids, all_sub_labels, all_seed_subs, all_obj_labels, all_lens, all_token_offsets, all_texts, all_spo_lists = batch

        X = self.X
        Y = self.Y
        Z = self.Z
        eval_loss = 0.0
        num_eval_steps = 0
        for i in range(all_input_ids.size()[0]):
            inputs = {
                "input_ids": all_input_ids[i].view(1, -1),
                "attention_mask": all_input_mask[i].view(1, -1),
                #  "sub_labels": all_sub_labels[i].view(1, -1),
                #  "seed_subs": all_seed_subs[i].view(1, -1),
                #  "obj_labels": all_obj_labels[i].view(1, -1),
                "do_forward_obj": False,
            }
            if args.model_type != "distilbert":
                # XLM and RoBERTa don"t use segment_ids
                inputs["token_type_ids"] = (all_segment_ids[i].view(
                    1, -1) if args.model_type in ["bert", "xlnet"] else None)
            token_offsets = all_token_offsets[i]
            text = all_texts[i]
            #  logger.warning(f"text: {len(text)} {text}")
            spo_list = all_spo_lists[i]

            outputs = model(**inputs)
            #  logger.warning(f"ouptuts: {outputs}")

            batch_eval_loss, subject_preds = outputs[:2]
            eval_loss += batch_eval_loss
            num_eval_steps += 1

            #  logger.info(f"subject_preds.shape: {subject_preds.shape}")
            subjects = extract_subjects_from_logits(args,
                                                    subject_preds,
                                                    confidence=args.confidence)
            #  logger.warning(f"text: {text}")
            #  logger.info(f"spo_list: {spo_list}")
            pred_subjects = []
            for sub in subjects:
                s_text = token_offsets[sub[0] - 1][0]
                e_text = token_offsets[sub[1] - 1][-1]
                pred_subjects.append(text[s_text:e_text])
            #  logger.info(f"pred_subjects: {pred_subjects}")
            pred_spoes = []
            if subjects:
                spoes = extract_predicate_objects_from_logits(
                    args,
                    model,
                    subjects,
                    inputs['input_ids'],
                    inputs['attention_mask'],
                    inputs['token_type_ids'],
                    token_offsets,
                    all_lens[i:i + 1],
                    confidence=args.confidence / 4)
                #  logger.info(f"Eval spoes: {spoes}")
                for spo in spoes:
                    (s_sub, e_sub), predicate, (s_obj, e_obj) = spo
                    s_sub_text = token_offsets[s_sub - 1][0].item()
                    e_sub_text = token_offsets[e_sub - 1][-1].item()
                    s_obj_text = token_offsets[s_obj - 1][0].item()
                    e_obj_text = token_offsets[e_obj - 1][-1].item()
                    if e_sub_text - s_sub_text >= 2 and e_obj_text - s_obj_text >= 2:
                        #  logger.info(f"{text[s_sub_text:e_sub_text]} , {args.id2label[predicate]}, {text[s_obj_text:e_obj_text]}")
                        pred_predicate = args.id2label[predicate]
                        s_mention = text[s_sub_text:e_sub_text]
                        pred_subject = (s_sub_text, s_mention)
                        o_mention = text[s_obj_text:e_obj_text]
                        pred_object = (s_obj_text, o_mention)
                        pred_spoes.append(
                            (pred_subject, pred_predicate, pred_object))
                #  logger.info(f"pred_spoes: {pred_spoes}")

                self.eval_results.append(spoes)
            x, y, z = evaluate_spo_list(pred_spoes, spo_list)
            X += x
            Y += y
            Z += z
            #  else:
            #      logger.warning(f"subjects is None.")

        eval_loss = eval_loss / num_eval_steps
        results = {}
        #  results = {f'{key}': value for key, value in eval_info.items()}
        results['loss'] = eval_loss
        f1, acc, recall = 2 * X / (Y + Z), X / Y, X / Z
        self.X = X
        self.Y = Y
        self.Z = Z
        results['f1'] = f1
        results['acc'] = acc
        results['recall'] = recall

        return (eval_loss, ), results

    #  def on_eval_step(self, args, eval_dataset, step, model, inputs, outputs):
    #  def on_eval_step_temp(self, args, model, step, batch, batch_features):
    #      all_input_ids, all_input_mask, all_segment_ids, all_sub_labels, all_seed_subs, all_obj_labels, all_lens = batch
    #
    #      eval_loss = 0.0
    #      num_eval_steps = 0
    #      for i in range(all_input_ids.size()[0]):
    #          inputs = {
    #              "input_ids": all_input_ids[i].view(1, -1),
    #              "attention_mask": all_input_mask[i].view(1, -1),
    #          }
    #          if args.model_type != "distilbert":
    #              # XLM and RoBERTa don"t use segment_ids
    #              inputs["token_type_ids"] = (all_segment_ids[i].view(
    #                  1, -1) if args.model_type in ["bert", "xlnet"] else None)
    #
    #          outputs = model(**inputs)
    #          tmp_eval_loss, logits = outputs[:2]
    #          eval_loss += tmp_eval_loss
    #          num_eval_steps += 1
    #
    #          #  loss_sig = nn.Sigmoid()
    #          #  active_logits = logits.view(-1, args.num_labels * 2)
    #          #  activae_logits = loss_sig(active_logits)
    #          #  active_logits = active_logits**2
    #          #  preds = active_logits
    #          preds = get_active_logits(logits, args.num_labels)
    #
    #          T = batch_features[i].subjects
    #          R = extract_entity_from_logits(
    #              args,
    #              preds,
    #              all_lens[i:i + 1],
    #              confidence=args.confidence,
    #              enable_nested_entities=args.enable_nested_entities)
    #          #  logger.info(f"R: {R}")
    #          #  logger.info(f"T: {T}")
    #
    #          self.metric.update(true_subject=T, pred_subject=R)
    #
    #      eval_loss = eval_loss / num_eval_steps
    #      eval_info, entity_info = self.metric.result()
    #      results = {f'{key}': value for key, value in eval_info.items()}
    #      results['loss'] = eval_loss
    #
    #      return (eval_loss, ), results

    def on_predict_start(self, args, test_dataset):
        self.pred_results = []

    def on_predict_step(self, args, model, step, batch):
        all_input_ids, all_input_mask, all_segment_ids, all_sub_labels, all_seed_subs, all_obj_labels, all_lens, all_token_offsets, all_texts, all_spo_lists = batch

        for i in range(all_input_ids.size()[0]):
            inputs = {
                "input_ids": all_input_ids[i].view(1, -1),
                "attention_mask": all_input_mask[i].view(1, -1),
                #  "sub_labels": all_sub_labels[i].view(1, -1),
                #  "seed_subs": all_seed_subs[i].view(1, -1),
                #  "obj_labels": all_obj_labels[i].view(1, -1),
                "do_forward_obj": False,
            }
            if args.model_type != "distilbert":
                # XLM and RoBERTa don"t use segment_ids
                inputs["token_type_ids"] = (all_segment_ids[i].view(
                    1, -1) if args.model_type in ["bert", "xlnet"] else None)
            token_offsets = all_token_offsets[i]
            text = all_texts[i]
            spo_list = all_spo_lists[i]

            outputs = model(**inputs)
            #  logger.warning(f"ouptuts: {outputs}")

            batch_eval_loss, subject_preds = outputs[:2]

            #  logger.info(f"subject_preds.shape: {subject_preds.shape}")
            subjects = extract_subjects_from_logits(args,
                                                    subject_preds,
                                                    confidence=args.confidence)
            #  logger.warning(f"text: {text}")
            #  logger.info(f"spo_list: {spo_list}")
            pred_subjects = []
            for sub in subjects:
                s_text = token_offsets[sub[0] - 1][0]
                e_text = token_offsets[sub[1] - 1][-1]
                pred_subjects.append(text[s_text:e_text])
            #  logger.info(f"pred_subjects: {pred_subjects}")
            pred_spoes = []
            if subjects:
                spoes = extract_predicate_objects_from_logits(
                    args,
                    model,
                    subjects,
                    inputs['input_ids'],
                    inputs['attention_mask'],
                    inputs['token_type_ids'],
                    token_offsets,
                    all_lens[i:i + 1],
                    confidence=args.confidence / 4)
                #  logger.info(f"Eval spoes: {spoes}")
                for spo in spoes:
                    (s_sub, e_sub), predicate, (s_obj, e_obj) = spo
                    s_sub_text = token_offsets[s_sub - 1][0].item()
                    e_sub_text = token_offsets[e_sub - 1][-1].item()
                    s_obj_text = token_offsets[s_obj - 1][0].item()
                    e_obj_text = token_offsets[e_obj - 1][-1].item()
                    if e_sub_text - s_sub_text >= 2 and e_obj_text - s_obj_text >= 2:
                        #  logger.info(f"{text[s_sub_text:e_sub_text]} , {args.id2label[predicate]}, {text[s_obj_text:e_obj_text]}")
                        pred_predicate = args.id2label[predicate]
                        s_mention = text[s_sub_text:e_sub_text]
                        pred_subject = (s_sub_text, s_mention)
                        o_mention = text[s_obj_text:e_obj_text]
                        pred_object = (s_obj_text, o_mention)
                        pred_spoes.append(
                            (pred_subject, pred_predicate, pred_object))
                #  logger.info(f"pred_spoes: {pred_spoes}")

            self.pred_results.append(pred_spoes)

    def on_predict_end(self, args, test_dataset):
        return self.pred_results

    def on_eval_end(self, args, eval_dataset):
        X = self.X
        Y = self.Y
        Z = self.Z
        f1, acc, recall = 2 * X / (Y + Z), X / Y, X / Z
        results = {}
        results['f1'] = f1
        results['acc'] = acc
        results['recall'] = recall
        return results



# -*- coding: utf-8 -*-
# File: __init__.py

# https://github.com/celery/kombu/blob/7d13f9b95d0b50c94393b962e6def928511bfda6/kombu/__init__.py#L34-L36
STATICA_HACK = True
globals()['kcah_acitats'[::-1].upper()] = False
if STATICA_HACK:
    from .base import *
    from .convert import *
    from .crop import *
    from .deform import *
    from .geometry import *
    from .imgproc import *
    from .meta import *
    from .misc import *
    from .noise import *
    from .paste import *
    from .transform import *
    from .external import *


import os
from pkgutil import iter_modules

__all__ = []


def global_import(name):
    p = __import__(name, globals(), locals(), level=1)
    lst = p.__all__ if '__all__' in dir(p) else dir(p)
    if lst:
        del globals()[name]
        for k in lst:
            if not k.startswith('__'):
                globals()[k] = p.__dict__[k]
                __all__.append(k)


try:
    import cv2  # noqa
except ImportError:
    from ...utils import logger
    logger.warn("Cannot import 'cv2', therefore image augmentation is not available.")
else:
    _CURR_DIR = os.path.dirname(__file__)
    for _, module_name, _ in iter_modules(
            [os.path.dirname(__file__)]):
        srcpath = os.path.join(_CURR_DIR, module_name + '.py')
        if not os.path.isfile(srcpath):
            continue
        if not module_name.startswith('_') and "_test" not in module_name:
            global_import(module_name)



# -*- coding: utf-8 -*-
# File: remote.py


import multiprocessing as mp
import time
from collections import deque
import tqdm

from ..utils import logger
from ..utils.concurrency import DIE
from ..utils.serialize import dumps, loads
from ..utils.utils import get_tqdm_kwargs
from .base import DataFlow, DataFlowReentrantGuard

try:
    import zmq
except ImportError:
    logger.warn("Error in 'import zmq'. remote feature won't be available")
    __all__ = []
else:
    __all__ = ['send_dataflow_zmq', 'RemoteDataZMQ']


def send_dataflow_zmq(df, addr, hwm=50, format=None, bind=False):
    """
    Run DataFlow and send data to a ZMQ socket addr.
    It will serialize and send each datapoint to this address with a PUSH socket.
    This function never returns.

    Args:
        df (DataFlow): Will infinitely loop over the DataFlow.
        addr: a ZMQ socket endpoint.
        hwm (int): ZMQ high-water mark (buffer size)
        format (str): The serialization format.
             Default format uses :mod:`utils.serialize`.
             This format works with :class:`dataflow.RemoteDataZMQ`.
             An alternate format is 'zmq_ops', used by https://github.com/tensorpack/zmq_ops
             and :class:`input_source.ZMQInput`.
        bind (bool): whether to bind or connect to the endpoint address.
    """
    assert format in [None, 'zmq_op', 'zmq_ops']
    if format is None:
        dump_fn = dumps
    else:
        from zmq_ops import dump_arrays
        dump_fn = dump_arrays

    ctx = zmq.Context()
    socket = ctx.socket(zmq.PUSH)
    socket.set_hwm(hwm)
    if bind:
        socket.bind(addr)
    else:
        socket.connect(addr)
    try:
        df.reset_state()
        logger.info("Serving data to {} with {} format ...".format(
            addr, 'default' if format is None else 'zmq_ops'))
        INTERVAL = 200
        q = deque(maxlen=INTERVAL)

        try:
            total = len(df)
        except NotImplementedError:
            total = 0
        tqdm_args = get_tqdm_kwargs(leave=True, smoothing=0.8)
        tqdm_args['bar_format'] = tqdm_args['bar_format'] + "{postfix}"
        while True:
            with tqdm.trange(total, **tqdm_args) as pbar:
                for dp in df:
                    start = time.time()
                    socket.send(dump_fn(dp), copy=False)
                    q.append(time.time() - start)
                    pbar.update(1)
                    if pbar.n % INTERVAL == 0:
                        avg = "{:.3f}".format(sum(q) / len(q))
                        pbar.set_postfix({'AvgSendLat': avg})
    finally:
        logger.info("Exiting send_dataflow_zmq ...")
        socket.setsockopt(zmq.LINGER, 0)
        socket.close()
        if not ctx.closed:
            ctx.destroy(0)


class RemoteDataZMQ(DataFlow):
    """
    Produce data from ZMQ PULL socket(s).
    It is the receiver-side counterpart of :func:`send_dataflow_zmq`, which uses :mod:`tensorpack.utils.serialize`
    for serialization.
    See http://tensorpack.readthedocs.io/tutorial/efficient-dataflow.html#distributed-dataflow

    Attributes:
        cnt1, cnt2 (int): number of data points received from addr1 and addr2
    """
    def __init__(self, addr1, addr2=None, hwm=50, bind=True):
        """
        Args:
            addr1,addr2 (str): addr of the zmq endpoint to connect to.
                Use both if you need two protocols (e.g. both IPC and TCP).
                I don't think you'll ever need 3.
            hwm (int): ZMQ high-water mark (buffer size)
            bind (bool): whether to connect or bind the endpoint
        """
        assert addr1
        self._addr1 = addr1
        self._addr2 = addr2
        self._hwm = int(hwm)
        self._guard = DataFlowReentrantGuard()
        self._bind = bind

    def reset_state(self):
        self.cnt1 = 0
        self.cnt2 = 0

    def bind_or_connect(self, socket, addr):
        if self._bind:
            socket.bind(addr)
        else:
            socket.connect(addr)

    def __iter__(self):
        with self._guard:
            try:
                ctx = zmq.Context()
                if self._addr2 is None:
                    socket = ctx.socket(zmq.PULL)
                    socket.set_hwm(self._hwm)
                    self.bind_or_connect(socket, self._addr1)

                    while True:
                        dp = loads(socket.recv(copy=False))
                        yield dp
                        self.cnt1 += 1
                else:
                    socket1 = ctx.socket(zmq.PULL)
                    socket1.set_hwm(self._hwm)
                    self.bind_or_connect(socket1, self._addr1)

                    socket2 = ctx.socket(zmq.PULL)
                    socket2.set_hwm(self._hwm)
                    self.bind_or_connect(socket2, self._addr2)

                    poller = zmq.Poller()
                    poller.register(socket1, zmq.POLLIN)
                    poller.register(socket2, zmq.POLLIN)

                    while True:
                        evts = poller.poll()
                        for sock, evt in evts:
                            dp = loads(sock.recv(copy=False))
                            yield dp
                            if sock == socket1:
                                self.cnt1 += 1
                            else:
                                self.cnt2 += 1
            finally:
                ctx.destroy(linger=0)


# for internal use only
def dump_dataflow_to_process_queue(df, size, nr_consumer):
    """
    Convert a DataFlow to a :class:`multiprocessing.Queue`.
    The DataFlow will only be reset in the spawned process.

    Args:
        df (DataFlow): the DataFlow to dump.
        size (int): size of the queue
        nr_consumer (int): number of consumer of the queue.
            The producer will add this many of ``DIE`` sentinel to the end of the queue.

    Returns:
        tuple(queue, process):
            The process will take data from ``df`` and fill
            the queue, once you start it. Each element in the queue is (idx,
            dp). idx can be the ``DIE`` sentinel when ``df`` is exhausted.
    """
    q = mp.Queue(size)

    class EnqueProc(mp.Process):

        def __init__(self, df, q, nr_consumer):
            super(EnqueProc, self).__init__()
            self.df = df
            self.q = q

        def run(self):
            self.df.reset_state()
            try:
                for idx, dp in enumerate(self.df):
                    self.q.put((idx, dp))
            finally:
                for _ in range(nr_consumer):
                    self.q.put((DIE, None))

    proc = EnqueProc(df, q, nr_consumer)
    return q, proc


if __name__ == '__main__':
    from argparse import ArgumentParser
    from .raw import FakeData
    from .common import TestDataSpeed

    """
    Test the multi-producer single-consumer model
    """
    parser = ArgumentParser()
    parser.add_argument('-t', '--task', choices=['send', 'recv'], required=True)
    parser.add_argument('-a', '--addr1', required=True)
    parser.add_argument('-b', '--addr2', default=None)
    args = parser.parse_args()

    # tcp addr like "tcp://127.0.0.1:8877"
    # ipc addr like "ipc://@ipc-test"
    if args.task == 'send':
        # use random=True to make it slow and cpu-consuming
        ds = FakeData([(128, 244, 244, 3)], 1000, random=True)
        send_dataflow_zmq(ds, args.addr1)
    else:
        ds = RemoteDataZMQ(args.addr1, args.addr2)
        logger.info("Each DP is 73.5MB")
        TestDataSpeed(ds).start_test()



# -*- coding: utf-8 -*-
# File: serialize.py

import os

import pickle
from multiprocessing.reduction import ForkingPickler
import msgpack
import msgpack_numpy

msgpack_numpy.patch()
assert msgpack.version >= (0, 5, 2)

__all__ = ['loads', 'dumps']


MAX_MSGPACK_LEN = 1000000000


class MsgpackSerializer(object):

    @staticmethod
    def dumps(obj):
        """
        Serialize an object.

        Returns:
            Implementation-dependent bytes-like object.
        """
        return msgpack.dumps(obj, use_bin_type=True)

    @staticmethod
    def loads(buf):
        """
        Args:
            buf: the output of `dumps`.
        """
        # Since 0.6, the default max size was set to 1MB.
        # We change it to approximately 1G.
        return msgpack.loads(buf, raw=False,
                             max_bin_len=MAX_MSGPACK_LEN,
                             max_array_len=MAX_MSGPACK_LEN,
                             max_map_len=MAX_MSGPACK_LEN,
                             max_str_len=MAX_MSGPACK_LEN)


class PyarrowSerializer(object):
    @staticmethod
    def dumps(obj):
        """
        Serialize an object.

        Returns:
            Implementation-dependent bytes-like object.
            May not be compatible across different versions of pyarrow.
        """
        import pyarrow as pa
        return pa.serialize(obj).to_buffer()

    @staticmethod
    def dumps_bytes(obj):
        """
        Returns:
            bytes
        """
        return PyarrowSerializer.dumps(obj).to_pybytes()

    @staticmethod
    def loads(buf):
        """
        Args:
            buf: the output of `dumps` or `dumps_bytes`.
        """
        import pyarrow as pa
        return pa.deserialize(buf)


class PickleSerializer(object):
    @staticmethod
    def dumps(obj):
        """
        Returns:
            bytes
        """
        return pickle.dumps(obj, protocol=-1)

    @staticmethod
    def loads(buf):
        """
        Args:
            bytes
        """
        return pickle.loads(buf)


# Define the default serializer to be used that dumps data to bytes
_DEFAULT_S = os.environ.get('TENSORPACK_SERIALIZE', 'pickle')

if _DEFAULT_S == "pyarrow":
    dumps = PyarrowSerializer.dumps_bytes
    loads = PyarrowSerializer.loads
elif _DEFAULT_S == "pickle":
    dumps = PickleSerializer.dumps
    loads = PickleSerializer.loads
else:
    dumps = MsgpackSerializer.dumps
    loads = MsgpackSerializer.loads

# Define the default serializer to be used for passing data
# among a pair of peers. In this case the deserialization is
# known to happen only once
_DEFAULT_S = os.environ.get('TENSORPACK_ONCE_SERIALIZE', 'pickle')

if _DEFAULT_S == "pyarrow":
    dumps_once = PyarrowSerializer.dumps
    loads_once = PyarrowSerializer.loads
elif _DEFAULT_S == "pickle":
    dumps_once = ForkingPickler.dumps
    loads_once = ForkingPickler.loads
else:
    dumps_once = MsgpackSerializer.dumps
    loads_once = MsgpackSerializer.loads



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import json
from copy import deepcopy
from dataclasses import dataclass, field
from typing import List

from loguru import logger
from tqdm import tqdm

#from .dataflow import RNGDataFlow


# 实体标签
@dataclass
class EntityTag:
    # 标签类别
    category: str = ""
    # 标签起始位置
    start: int = -1
    # 标签文本
    mention: str = ""

    def clear(self):
        pass

    def to_dict(self):
        return {
            'category': self.category,
            'start': self.start,
            'mention': self.mention
        }

    def from_dict(self, dict_data):
        self.clear()
        for k in self.__dict__.keys():
            if k in dict_data:
                v = dict_data[k]
                setattr(self, k, v)
        return self

    def __repr__(self):
        return f"{self.to_dict()}"


# 打过标签的文本
@dataclass
class TaggedText:
    guid: str
    text: str
    tags: List[EntityTag] = field(default_factory=list)
    text_offset: int = 0

    def clear(self):
        self.tags = []

    def offset(self, text_offset):
        self.text_offset += text_offset
        for tag in self.tags:
            tag.start += text_offset

    def to_dict(self):
        return {
            'guid': self.guid,
            'text': self.text,
            'tags': [x.to_dict() for x in self.tags]
        }

    def from_dict(self, dict_data):
        self.clear()
        if 'guid' in dict_data:
            self.guid = dict_data['guid']
        elif 'text' in dict_data:
            self.text = dict_data['text']
        elif 'tags' in dict_data:
            for x in dict_data['tags']:
                self.tags.append(EntityTag().from_dict(x))

        return self

    def add_tag(self, tag: EntityTag):
        self.tags.append(tag)

    def __repr__(self):
        #  return f"{self.guid}: {self.text[:50]} | {self.text_offset} | {self.tags}"
        return f"guid: {self.guid}, text: {self.text[:50]}... | text_offset: {self.text_offset} | tags: {self.tags}"


#class BaseDataFlow(RNGDataFlow):
class BaseDataFlow(object):
    def __init__(self, data_list=None):
        super(BaseDataFlow, self).__init__()

        self._data_list = data_list if data_list is not None else []
        self._shuffle_idxs = []

    def __len__(self):
        return len(self._data_list)

    def __iter__(self):
        if len(self._shuffle_idxs) != self.__len__():
            self._shuffle_idxs = list(range(self.__len__()))

        for k in self._shuffle_idxs:
            yield self._data_list[k]

    def shuffle(self):
        self._shuffle_idxs = list(range(self.__len__()))
        #self.rng.shuffle(self._shuffle_idxs)

    def __getitem__(self, idx):
        return self._data_list[idx]

    def reset_state(self):
        #super(BaseDataFlow, self).reset_state()
        pass

    def clean(self):
        self._data_list = []
        self.reset_state()

    def append(self, data):
        self._data_list.append(data)

    def extend(self, other_dataset):
        self._data_list.extend(other_dataset._data_list)


class EntityDataFlow(BaseDataFlow):
    def __init__(self,
                 data_file=None,
                 data_generator=None,
                 data_list=None,
                 entity_labels=None):
        super(EntityDataFlow, self).__init__(data_list=data_list)

        self.entity_labels = entity_labels
        self.data_generator = data_generator
        self.data_file = data_file

        if data_list is None:
            if self.data_generator is not None or self.data_file is not None:
                self.load()

    def info(self):
        logger.info(f"Entity labels: {self.entity_labels}")
        logger.info(f"data_file: {self.data_file}")
        logger.info(f"Total: {self.__len__()}")
        for i in range(5):
            logger.info(f"{self.__getitem__(i)}")

    # ------------------------------ Load ------------------------------
    def load(self):
        if self.data_generator is not None:
            self.load_from_generator(self.data_generator)
        elif self.data_file is not None:
            self.load_from_file(self.data_file)
        else:
            raise ValueError(
                f"Either data_generator or data_file must be set.")

    def load_from_generator(self, data_generator, data_source=None):
        self.clean()

        self.data_generator = data_generator

        categories = []

        g = self.data_generator(data_source) if callable(
            self.data_generator) else self.data_generator
        for guid, text, _, json_tags in g:
            tagged_text = TaggedText(guid, text)
            if json_tags:
                for json_tag in json_tags:
                    entity_tag = EntityTag().from_dict(json_tag)
                    tagged_text.add_tag(entity_tag)

                    category = entity_tag.category
                    if category not in categories:
                        categories.append(category)

            self.append(tagged_text)

        logger.info(f"Loaded categories: {categories}")
        logger.info(f"Loaded {self.__len__()} samples.")
        if not self.entity_labels:
            self.entity_labels = categories

        return self

    def load_from_file(self, data_file: str):
        self.clean()

        self.data_file = data_file

        logger.info(f"Loading {data_file}")
        try:
            lines = [json.loads(x) for x in open(data_file, 'r').readlines()]
        except:
            logger.warning(
                f"{data_file} is not standard format, try reviews format...")
            try:
                json_data = json.load(open(data_file, 'r'))
                lines = [v for k, v in json_data.items()]
            except:
                logger.error(f"Unknown file format {data_file}")
                raise ValueError(f"Unknown file format {data_file}")

        def data_generator(data_source=None):
            for json_data in lines:
                yield json_data['guid'], json_data['text'], None, json_data[
                    'tags']

        return self.load_from_generator(data_generator)

    # ------------------------------ Save ------------------------------
    def save(self, filename: str, format="json"):
        """
        {'guid': '0000', 'text': "sample text", 'tags': [{'category': "实体类别1", start: 10, mention: "片断文本"}, ...]}
        """

        if format == "json":
            json_data = {x.guid: x.to_dict() for x in self}
            json.dump(json_data,
                      open(filename, 'w'),
                      ensure_ascii=False,
                      indent=2)
        elif format == 'lines':
            with open(filename, 'w') as wt:
                for tagged_text in tqdm(self, desc="Save tagged text list."):
                    data_dict = tagged_text.to_dict()
                    wt.write(f"{json.dumps(data_dict, ensure_ascii=False)}\n")
        else:
            raise ValueError(
                f"Bad format: {format}. Format must be one of ['json', 'lines']"
            )
        logger.warning(f"Saved {filename}")

    # ------------------------------ Brat data ------------------------------
    def import_from_brat_data(self, brat_data_dir):
        from .brat import brat_data_generator, get_brat_schemas
        entity_labels, relation_labels = get_brat_schemas(brat_data_dir)
        self.entity_labels = entity_labels

        return self.load_from_generator(brat_data_generator, brat_data_dir)

    def export_to_brat_data(self, brat_data_dir, max_pages=10):
        from .brat import export_brat_files
        export_brat_files(self._data_list,
                          self.entity_labels,
                          None,
                          brat_data_dir,
                          max_pages=max_pages)

    # ------------------------------ Poplar data ------------------------------
    def import_from_poplar_data(self, poplar_file):
        from .poplar import poplar_data_generator
        return self.load_from_generator(poplar_data_generator, poplar_file)

    def export_to_poplar_data(self, poplar_file, max_pages=100, start_page=0):
        from .poplar import save_poplar_file
        save_poplar_file(self._data_list,
                         poplar_file,
                         self.entity_labels,
                         None,
                         start_page=start_page,
                         max_pages=max_pages)

    # ------------------------------ Utils ------------------------------
    def diff(self, another_dataflow):
        assert len(self) == len(another_dataflow)

        identical_tags_list = []
        a_only_tags_list = []
        b_only_tags_list = []
        for tagged_text_a, tagged_text_b in zip(self, another_dataflow):
            text_a, text_b = tagged_text_a.text, tagged_text_b.text
            # 被标注的文本必须相同，否则标注位置无意义
            assert text_a == text_b
            tags_a, tags_b = tagged_text_a.tags, tagged_text_b.tags
            tags_a = sorted(tags_a, key=lambda x: x.start)
            tags_b = sorted(tags_b, key=lambda x: x.start)

            identical_tags = []
            a_only_tags = []
            b_only_tags = []
            for tag_a in tags_a:
                found_a = False
                for tag_b in tags_b:
                    if tag_a == tag_b:
                        identical_tags.append(tag_a)
                        found_a = True
                        break
                if not found_a:
                    a_only_tags.append(tag_a)
            for tag_b in tags_b:
                found_b = False
                for tag_a in tags_a:
                    if tag_a == tag_b:
                        found_b = True
                        break
                if not found_b:
                    b_only_tags.append(tag_b)

            identical_tags_list.append(identical_tags)
            a_only_tags_list.append(a_only_tags)
            b_only_tags_list.append(b_only_tags)

        return identical_tags_list, a_only_tags_list, b_only_tags_list


def merge_entity_dataflows(entity_dataflow_list, min_dups=2):
    assert entity_dataflow_list is not None
    assert len(entity_dataflow_list) > 0

    def remove_duplicate_entities(R):
        new_R = []
        for i, r in enumerate(R):
            found_dup = False
            for r1 in R[i + 1:]:
                if r == r1:
                    found_dup = True
                    break
            if found_dup:
                continue
            new_R.append(r)
        return new_R

    def is_uniform(entities):
        for i, x in enumerate(entities):
            for x1 in entities[i:]:
                if x != x1:
                    return False
        return True

    def merge_entities(entities_list, key=None, min_dups=2):
        new_X = []
        X = [x for z in entities_list for x in z]
        #  logger.info(f"X: {X}")
        if key:
            X = sorted(X, key=key)

        for i, x in enumerate(X[:1 - min_dups]):
            if is_uniform(X[i:i + min_dups]):
                new_X.append(deepcopy(x))
        new_X = remove_duplicate_entities(new_X)
        return new_X

    merged_entity_dataflow = deepcopy(entity_dataflow_list[0])
    if len(entity_dataflow_list) > 1:
        for i, X in enumerate(
                tqdm(zip(*entity_dataflow_list),
                     desc="Merge entity dataflows")):
            tags_list = [x.tags for x in X]
            new_tags = merge_entities(
                tags_list,
                key=lambda x: x.start * 1000 + len(x.mention),
                min_dups=min_dups)
            merged_entity_dataflow[i].tags = new_tags

    return merged_entity_dataflow


def mix_entity_dataflows(entity_dataflow_list):
    """
    以entity_dataflow_list[0]为基准，只保留实体文本区间不重叠的部分
    """
    if len(entity_dataflow_list) == 0:
        return None
    mixed_entity_dataflow = deepcopy(entity_dataflow_list[0])
    for i, X in enumerate(
            tqdm(zip(*entity_dataflow_list), desc="Mix ner datasets")):
        mixed_tags = mixed_entity_dataflow[i].tags
        #  logger.info(f"mixed_tags: {mixed_tags}")
        tags_list = [x.tags for x in X]
        for tags in tags_list:
            for tag in tags:
                c = tag.category
                s = tag.start
                m = tag.mention
                e = s + len(m) - 1
                overlap = False
                for tag0 in mixed_tags:
                    #  logger.info(f"tag0: {tag0}")
                    c0 = tag0.category
                    s0 = tag0.start
                    m0 = tag0.mention
                    e0 = s0 + len(m0) - 1
                    if s >= s0 and s <= e0:
                        overlap = True
                        break
                    if e >= s0 and e <= e0:
                        overlap = True
                        break
                    if s0 >= s and s0 <= e:
                        overlap = True
                        break
                    if e0 >= s and e0 <= e:
                        overlap = True
                        break
                if not overlap:
                    mixed_tags.append(EntityTag(c, s, m))
        mixed_tags = sorted(mixed_tags, key=lambda x: x.start)
        mixed_entity_dataflow[i].tags = mixed_tags
    return mixed_entity_dataflow


if __name__ == '__main__':
    # from theta.nlp.dataflow import EntityDataFlow
    df = EntityDataFlow(
        data_file="../../tutorials/cluener/cluener_train_data.json")
    df.info()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# ----------------------------------------------------------
#  from typing import Tuple
#  import torch
#  import torch.nn as nn
#  from torch import nn, Tensor
#  import torch.nn.functional as F
#
#  class CircleLoss(nn.Module):
#      def __init__(self, gamma=32, m=0.25, similarity='cos', **kwargs):
#          super(CircleLoss, self).__init__()
#          self.gamma = gamma
#          self.m = m
#          self.similarity = similarity
#
#      def forward(self, feats, labels):
#          assert feats.size(0) == labels.size(0), \
#              f"feats.size(0): {feats.size(0)} is not equal to labels.size(0): {labels.size(0)}"
#
#          m = labels.size(0)
#          mask = labels.expand(m, m).t().eq(labels.expand(m, m)).float()
#          pos_mask = mask.triu(diagonal=1)
#          neg_mask = (mask - 1).abs_().triu(diagonal=1)
#          if self.similarity == 'dot':
#              sim_mat = torch.matmul(feats, torch.t(feats))
#          elif self.similarity == 'cos':
#              feats = F.normalize(feats)
#              sim_mat = feats.mm(feats.t())
#          else:
#              raise ValueError('This similarity is not implemented.')
#
#          pos_pair_ = sim_mat[pos_mask == 1]
#          neg_pair_ = sim_mat[neg_mask == 1]
#
#          alpha_p = torch.relu(-pos_pair_ + 1 + self.m)
#          alpha_n = torch.relu(neg_pair_ + self.m)
#          margin_p = 1 - self.m
#          margin_n = self.m
#          loss_p = torch.sum(
#              torch.exp(-self.gamma * alpha_p * (pos_pair_ - margin_p)))
#          loss_n = torch.sum(
#              torch.exp(self.gamma * alpha_n * (neg_pair_ - margin_n)))
#          loss = torch.log(1 + loss_p * loss_n)
#          return loss

# -------------------------------------------------------
import torch
import torch.nn as nn
import torch.nn.functional as F


def feature_similarity(x, label, similarity='cos'):
    n = label.size(0)
    mask = label.expand(n, n).t().eq(label.expand(n, n)).float()
    pos_mask = mask.triu(diagonal=1)
    neg_mask = (mask - 1).abs_().triu(diagonal=1)

    # inspired by attention
    x = x.squeeze()
    if similarity == 'dot':
        sim_mat = x.matmul(x.transpose(0, 1))
    elif similarity == 'cos':
        x = F.normalize(x)
        sim_mat = x.matmul(x.transpose(0, 1))

    return sim_mat[pos_mask == 1], sim_mat[neg_mask == 1]


class CircleLoss(nn.Module):
    def __init__(self, gamma=80, m=0.25, similarity='cos'):
        super(CircleLoss, self).__init__()
        self.m = m
        self.gamma = gamma
        self.similarity = similarity
        self.soft_plus = nn.Softplus()

    def forward(self, x, label):
        sp, sn = feature_similarity(x, label, similarity=self.similarity)
        ap = torch.relu(-sp.detach() + 1 + self.m)
        an = torch.relu(sn.detach() + self.m)

        delta_p = 1 - self.m
        delta_n = self.m

        logit_p = -ap * (sp - delta_p) * self.gamma
        logit_n = an * (sn - delta_n) * self.gamma

        loss_p = torch.sum(torch.exp(logit_p))
        loss_n = torch.sum(torch.exp(logit_n))

        loss = torch.log(1 + loss_p * loss_n)

        return loss


# -------------------------------------------------------
#  from typing import Tuple
#
#  import torch
#  from torch import nn, Tensor
#
#
#  def convert_label_to_similarity(normed_feature: Tensor,
#                                  label: Tensor) -> Tuple[Tensor, Tensor]:
#      similarity_matrix = normed_feature @ normed_feature.transpose(1, 0)
#      label_matrix = label.unsqueeze(1) == label.unsqueeze(0)
#
#      positive_matrix = label_matrix.triu(diagonal=1)
#      negative_matrix = label_matrix.logical_not().triu(diagonal=1)
#
#      similarity_matrix = similarity_matrix.view(-1)
#      positive_matrix = positive_matrix.view(-1)
#      negative_matrix = negative_matrix.view(-1)
#      return similarity_matrix[positive_matrix], similarity_matrix[
#          negative_matrix]
#
#
#  class CircleLoss(nn.Module):
#      #  def __init__(self, m: float, gamma: float) -> None:
#      def __init__(self, gamma: float = 32, m: float = 0.25) -> None:
#          super(CircleLoss, self).__init__()
#          self.m = m
#          self.gamma = gamma
#          self.soft_plus = nn.Softplus()
#
#      def forward(self, sp: Tensor, sn: Tensor) -> Tensor:
#          ap = torch.clamp_min(-sp.detach() + 1 + self.m, min=0.)
#          an = torch.clamp_min(sn.detach() + self.m, min=0.)
#
#          delta_p = 1 - self.m
#          delta_n = self.m
#
#          logit_p = -ap * (sp - delta_p) * self.gamma
#          logit_n = an * (sn - delta_n) * self.gamma
#
#          loss = self.soft_plus(
#              torch.logsumexp(logit_n, dim=0) + torch.logsumexp(logit_p, dim=0))
#
#          return loss
#
#
#  if __name__ == "__main__":
#      feat = nn.functional.normalize(torch.rand(256, 64, requires_grad=True))
#      lbl = torch.randint(high=10, size=(256, ))
#
#      inp_sp, inp_sn = convert_label_to_similarity(feat, lbl)
#
#      criterion = CircleLoss(m=0.25, gamma=256)
#      circle_loss = criterion(inp_sp, inp_sn)
#
#      print(circle_loss)
#
#  #  def convert_label_to_similarity(normed_feature: Tensor, label: Tensor) -> Tuple[Tensor, Tensor]:
#  #      similarity_matrix = normed_feature @ normed_feature.transpose(1, 0)
#  #      label_matrix = label.unsqueeze(1) == label.unsqueeze(0)
#  #
#  #      positive_matrix = label_matrix.triu(diagonal=1)
#  #      negative_matrix = label_matrix.logical_not().triu(diagonal=1)
#  #
#  #      similarity_matrix = similarity_matrix.view(-1)
#  #      positive_matrix = positive_matrix.view(-1)
#  #      negative_matrix = negative_matrix.view(-1)
#  #      return similarity_matrix[positive_matrix], similarity_matrix[negative_matrix]
#  #
#  #
#  #  class CircleLoss(nn.Module):
#  #      def __init__(self, gamma: float, m: float) -> None:
#  #          super(CircleLoss, self).__init__()
#  #          self.m = m
#  #          self.gamma = gamma
#  #          self.soft_plus = nn.Softplus()
#  #
#  #      #def forward(self, sp: Tensor, sn: Tensor) -> Tensor:
#  #  def forward(self, feats, labels):
#  #      sp, sn = convert_label_to_similarity(feats, labels)
#  #          sp = nn.functional.normalize(sp)
#  #          sp, sn = convert_label_to_similarity(sp, sn)
#  #          ap = torch.clamp_min(- sp.detach() + 1 + self.m, min=0.)
#  #          an = torch.clamp_min(sn.detach() + self.m, min=0.)
#  #
#  #          delta_p = 1 - self.m
#  #          delta_n = self.m
#  #
#  #          logit_p = - ap * (sp - delta_p) * self.gamma
#  #          logit_n = an * (sn - delta_n) * self.gamma
#  #
#  #          loss = self.soft_plus(torch.logsumexp(logit_n, dim=0) + torch.logsumexp(logit_p, dim=0))
#  #
#  #          return loss



# -*- coding: utf-8 -*-
# File: concurrency.py

# Some code taken from zxytim

import sys
import atexit
import bisect
import multiprocessing as mp
import platform
import signal
import threading
import weakref
from contextlib import contextmanager
import six
from six.moves import queue
import subprocess

from . import logger
from .argtools import log_once


__all__ = ['StoppableThread', 'LoopThread', 'ShareSessionThread',
           'ensure_proc_terminate',
           'start_proc_mask_signal']


class StoppableThread(threading.Thread):
    """
    A thread that has a 'stop' event.
    """

    def __init__(self, evt=None):
        """
        Args:
            evt(threading.Event): if None, will create one.
        """
        super(StoppableThread, self).__init__()
        if evt is None:
            evt = threading.Event()
        self._stop_evt = evt

    def stop(self):
        """ Stop the thread"""
        self._stop_evt.set()

    def stopped(self):
        """
        Returns:
            bool: whether the thread is stopped or not
        """
        return self._stop_evt.isSet()

    def queue_put_stoppable(self, q, obj):
        """ Put obj to queue, but will give up when the thread is stopped"""
        while not self.stopped():
            try:
                q.put(obj, timeout=5)
                break
            except queue.Full:
                pass

    def queue_get_stoppable(self, q):
        """ Take obj from queue, but will give up when the thread is stopped"""
        while not self.stopped():
            try:
                return q.get(timeout=5)
            except queue.Empty:
                pass


class LoopThread(StoppableThread):
    """ A pausable thread that simply runs a loop"""

    def __init__(self, func, pausable=True):
        """
        Args:
            func: the function to run
        """
        super(LoopThread, self).__init__()
        self._func = func
        self._pausable = pausable
        if pausable:
            self._lock = threading.Lock()
        self.daemon = True

    def run(self):
        while not self.stopped():
            if self._pausable:
                self._lock.acquire()
                self._lock.release()
            self._func()

    def pause(self):
        """ Pause the loop """
        assert self._pausable
        self._lock.acquire()

    def resume(self):
        """ Resume the loop """
        assert self._pausable
        self._lock.release()


class ShareSessionThread(threading.Thread):
    """ A wrapper around thread so that the thread
        uses the default session at "start()" time.
    """
    def __init__(self, th=None):
        """
        Args:
            th (threading.Thread or None):
        """
        super(ShareSessionThread, self).__init__()
        if th is not None:
            assert isinstance(th, threading.Thread), th
            self._th = th
            self.name = th.name
            self.daemon = th.daemon

    @contextmanager
    def default_sess(self):
        if self._sess:
            with self._sess.as_default():
                yield self._sess
        else:
            logger.warn("ShareSessionThread {} wasn't under a default session!".format(self.name))
            yield None

    def start(self):
        from ..compat import tfv1
        self._sess = tfv1.get_default_session()
        super(ShareSessionThread, self).start()

    def run(self):
        if not self._th:
            raise NotImplementedError()
        with self._sess.as_default():
            self._th.run()


class DIE(object):
    """ A placeholder class indicating end of queue """
    pass


def ensure_proc_terminate(proc):
    """
    Make sure processes terminate when main process exit.

    Args:
        proc (multiprocessing.Process or list)
    """
    if isinstance(proc, list):
        for p in proc:
            ensure_proc_terminate(p)
        return

    def stop_proc_by_weak_ref(ref):
        proc = ref()
        if proc is None:
            return
        if not proc.is_alive():
            return
        proc.terminate()
        proc.join()

    assert isinstance(proc, mp.Process)
    atexit.register(stop_proc_by_weak_ref, weakref.ref(proc))


def enable_death_signal(_warn=True):
    """
    Set the "death signal" of the current process, so that
    the current process will be cleaned with guarantee
    in case the parent dies accidentally.
    """
    if platform.system() != 'Linux':
        return
    try:
        import prctl    # pip install python-prctl
    except ImportError:
        if _warn:
            log_once('"import prctl" failed! Install python-prctl so that processes can be cleaned with guarantee.',
                     'warn')
        return
    else:
        assert hasattr(prctl, 'set_pdeathsig'), \
            "prctl.set_pdeathsig does not exist! Note that you need to install 'python-prctl' instead of 'prctl'."
        # is SIGHUP a good choice?
        prctl.set_pdeathsig(signal.SIGHUP)


def is_main_thread():
    if six.PY2:
        return isinstance(threading.current_thread(), threading._MainThread)
    else:
        # a nicer solution with py3
        return threading.current_thread() == threading.main_thread()


@contextmanager
def mask_sigint():
    """
    Returns:
        If called in main thread, returns a context where ``SIGINT`` is ignored, and yield True.
        Otherwise yield False.
    """
    if is_main_thread():
        sigint_handler = signal.signal(signal.SIGINT, signal.SIG_IGN)
        yield True
        signal.signal(signal.SIGINT, sigint_handler)
    else:
        yield False


def start_proc_mask_signal(proc):
    """
    Start process(es) with SIGINT ignored.

    Args:
        proc: (mp.Process or list)

    Note:
        The signal mask is only applied when called from main thread.
    """
    if not isinstance(proc, list):
        proc = [proc]

    with mask_sigint():
        for p in proc:
            if isinstance(p, mp.Process):
                if sys.version_info < (3, 4) or mp.get_start_method() == 'fork':
                    log_once("""
Starting a process with 'fork' method is efficient but not safe and may cause deadlock or crash.
Use 'forkserver' or 'spawn' method instead if you run into such issues.
See https://docs.python.org/3/library/multiprocessing.html#contexts-and-start-methods on how to set them.
""".replace("\n", ""),
'warn')  # noqa
            p.start()


def subproc_call(cmd, timeout=None):
    """
    Execute a command with timeout, and return STDOUT and STDERR

    Args:
        cmd(str): the command to execute.
        timeout(float): timeout in seconds.

    Returns:
        output(bytes), retcode(int). If timeout, retcode is -1.
    """
    try:
        output = subprocess.check_output(
            cmd, stderr=subprocess.STDOUT,
            shell=True, timeout=timeout)
        return output, 0
    except subprocess.TimeoutExpired as e:
        logger.warn("Command '{}' timeout!".format(cmd))
        if e.output:
            logger.warn(e.output.decode('utf-8'))
            return e.output, -1
        else:
            return "", -1
    except subprocess.CalledProcessError as e:
        logger.warn("Command '{}' failed, return code={}".format(cmd, e.returncode))
        logger.warn(e.output.decode('utf-8'))
        return e.output, e.returncode
    except Exception:
        logger.warn("Command '{}' failed to run.".format(cmd))
        return "", -2


class OrderedContainer(object):
    """
    Like a queue, but will always wait to receive item with rank
    (x+1) and produce (x+1) before producing (x+2).

    Warning:
        It is not thread-safe.
    """

    def __init__(self, start=0):
        """
        Args:
            start(int): the starting rank.
        """
        self.ranks = []
        self.data = []
        self.wait_for = start

    def put(self, rank, val):
        """
        Args:
            rank(int): rank of th element. All elements must have different ranks.
            val: an object
        """
        idx = bisect.bisect(self.ranks, rank)
        self.ranks.insert(idx, rank)
        self.data.insert(idx, val)

    def has_next(self):
        if len(self.ranks) == 0:
            return False
        return self.ranks[0] == self.wait_for

    def get(self):
        assert self.has_next()
        ret = self.data[0]
        rank = self.ranks[0]
        del self.ranks[0]
        del self.data[0]
        self.wait_for += 1
        return rank, ret


class OrderedResultGatherProc(mp.Process):
    """
    Gather indexed data from a data queue, and produce results with the
    original index-based order.
    """

    def __init__(self, data_queue, nr_producer, start=0):
        """
        Args:
            data_queue(mp.Queue): a queue which contains datapoints.
            nr_producer(int): number of producer processes. This process will
                terminate after receiving this many of :class:`DIE` sentinel.
            start(int): the rank of the first object
        """
        super(OrderedResultGatherProc, self).__init__()
        self.data_queue = data_queue
        self.ordered_container = OrderedContainer(start=start)
        self.result_queue = mp.Queue()
        self.nr_producer = nr_producer

    def run(self):
        nr_end = 0
        try:
            while True:
                task_id, data = self.data_queue.get()
                if task_id == DIE:
                    self.result_queue.put((task_id, data))
                    nr_end += 1
                    if nr_end == self.nr_producer:
                        return
                else:
                    self.ordered_container.put(task_id, data)
                    while self.ordered_container.has_next():
                        self.result_queue.put(self.ordered_container.get())
        except Exception as e:
            import traceback
            traceback.print_exc()
            raise e

    def get(self):
        return self.result_queue.get()



#! /usr/bin/env python
# -*- coding: utf-8 -*-

import tempfile
import numpy as np
import os
import unittest

from tensorpack.dataflow import HDF5Serializer, LMDBSerializer, NumpySerializer, TFRecordSerializer
from tensorpack.dataflow.base import DataFlow


def delete_file_if_exists(fn):
    try:
        os.remove(fn)
    except OSError:
        pass


class SeededFakeDataFlow(DataFlow):
    """docstring for SeededFakeDataFlow"""

    def __init__(self, seed=42, size=32):
        super(SeededFakeDataFlow, self).__init__()
        self.seed = seed
        self._size = size
        self.cache = []

    def reset_state(self):
        np.random.seed(self.seed)
        for _ in range(self._size):
            label = np.random.randint(low=0, high=10)
            img = np.random.randn(28, 28, 3)
            self.cache.append([label, img])

    def __len__(self):
        return self._size

    def __iter__(self):
        for dp in self.cache:
            yield dp


class SerializerTest(unittest.TestCase):

    def run_write_read_test(self, file, serializer, w_args, w_kwargs, r_args, r_kwargs, error_msg):
        try:
            delete_file_if_exists(file)

            ds_expected = SeededFakeDataFlow()
            serializer.save(ds_expected, file, *w_args, **w_kwargs)
            ds_actual = serializer.load(file, *r_args, **r_kwargs)

            ds_actual.reset_state()
            ds_expected.reset_state()

            for dp_expected, dp_actual in zip(ds_expected.__iter__(), ds_actual.__iter__()):
                self.assertEqual(dp_expected[0], dp_actual[0])
                self.assertTrue(np.allclose(dp_expected[1], dp_actual[1]))
        except ImportError:
            print(error_msg)

    def test_lmdb(self):
        with tempfile.TemporaryDirectory() as f:
            self.run_write_read_test(
                os.path.join(f, 'test.lmdb'),
                LMDBSerializer,
                {}, {},
                {}, {'shuffle': False},
                'Skip test_lmdb, no lmdb available')

    def test_tfrecord(self):
        with tempfile.TemporaryDirectory() as f:
            self.run_write_read_test(
                os.path.join(f, 'test.tfrecord'),
                TFRecordSerializer,
                {}, {},
                {}, {'size': 32},
                'Skip test_tfrecord, no tensorflow available')

    def test_numpy(self):
        with tempfile.TemporaryDirectory() as f:
            self.run_write_read_test(
                os.path.join(f, 'test.npz'),
                NumpySerializer,
                {}, {},
                {}, {'shuffle': False},
                'Skip test_numpy, no numpy available')

    def test_hdf5(self):
        args = [['label', 'image']]
        with tempfile.TemporaryDirectory() as f:
            self.run_write_read_test(
                os.path.join(f, 'test.h5'),
                HDF5Serializer,
                args, {},
                args, {'shuffle': False},
                'Skip test_hdf5, no h5py available')


if __name__ == '__main__':
    unittest.main()
